import argparse

from easy_fuse.ref_data.gff3_to_db import add_gff3_to_db_args
from easy_fuse.ref_data.gtf2tsl import add_gtf_to_tsl_args
from easy_fuse.misc.qc_parser import add_qc_parser_args
from easy_fuse.read_selection import add_read_selection_args
from easy_fuse.requantify import add_requantify_args
from easy_fuse.summarize_data import add_summarize_data_args
from easy_fuse.fusionreadfilter import add_read_filter_args
from easy_fuse.fusiontoolparser import add_fusion_parse_args
from easy_fuse.fusionannotation import add_annotation_parser_args
from logzero import logger
import easy_fuse
from easy_fuse.processing import add_pipeline_parser_args
from easy_fuse.tool_wrapper.skewer_wrapper import add_skewer_args
from easy_fuse.tool_wrapper.soapfuse_wrapper import add_soapfuse_wrapper_args
from easy_fuse.tool_wrapper.star_custom_index import add_star_custom_index_args


epilog = "Copyright (c) 2023 TRON gGmbH (See LICENSE for licensing details)"


def easy_fuse_cli():
    # set up logger
    parser = argparse.ArgumentParser(
        description="EasyFuse v{}".format(easy_fuse.__version__),
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        epilog=epilog,
    )

    subparsers = parser.add_subparsers(description="Commands")

    pipeline_parser = subparsers.add_parser(
        "pipeline", description="Runs the complete EasyFuse pipeline", epilog=epilog
    )
    add_pipeline_parser_args(pipeline_parser)

    gff3_to_db_parser = subparsers.add_parser(
        "gff3-to-db",
        description="Converts GFF3 to DB format",
        epilog=epilog,
    )
    add_gff3_to_db_args(gff3_to_db_parser)

    gtf_to_tsl_parser = subparsers.add_parser(
        "gtf-to-tsl",
        description="Converts GTF to TSL format",
        epilog=epilog,
    )
    add_gtf_to_tsl_args(gtf_to_tsl_parser)

    
    qc_parser_parser = subparsers.add_parser(
        "qc-parser",
        description="Parses QC data",
        epilog=epilog,
    )
    add_qc_parser_args(qc_parser_parser)

    skewer_wrapper_parser = subparsers.add_parser(
        "skewer-wrapper",
        description="Runs skewer",
        epilog=epilog,
    )
    add_skewer_args(skewer_wrapper_parser)

    read_filter_parser = subparsers.add_parser(
        "read-filter",
        description="Filter wild-type reads before fusion detection",
        epilog=epilog,
    )
    add_read_filter_args(read_filter_parser)

    soapfuse_wrapper_parser = subparsers.add_parser(
        "soapfuse-wrapper",
        description="Runs soapfuse",
        epilog=epilog,
    )
    add_soapfuse_wrapper_args(soapfuse_wrapper_parser)

    fusion_parser = subparsers.add_parser(
        "fusion-parser",
        description="Parse predicted gene fusions from tools into common breakpoint format",
        epilog=epilog,
    )
    add_fusion_parse_args(fusion_parser)

    annotation_parser = subparsers.add_parser(
        "annotation",
        description="Annotates transcript-level breakpoints based on gene model",
        epilog=epilog,
    )
    add_annotation_parser_args(annotation_parser)

    star_index_parser = subparsers.add_parser(
        "star-index",
        description="Runs a custom STAR index",
        epilog=epilog,
    )
    add_star_custom_index_args(star_index_parser)

    requantify_parser = subparsers.add_parser(
        "requantify",
        description="Counts junction and spanning reads from targeted STAR mapping (requantification)",
        epilog=epilog,
    )
    add_requantify_args(requantify_parser)

    read_selection_parser = subparsers.add_parser(
        "requantify-filter",
        description="Requantificaiton with more strict read filtering",
        epilog=epilog,
    )
    add_read_selection_args(read_selection_parser)

    summarize_data_parser = subparsers.add_parser(
        "summarize-data",
        description="Collect anotations and read counts from fusion detection tools and re-quantification. "
                    "Calls prediction model",
        epilog=epilog,
    )
    add_summarize_data_args(summarize_data_parser)

    args = parser.parse_args()

    try:
        args.func(args)
    except AttributeError as e:
        logger.exception(e)
        parser.parse_args(["--help"])



#!/usr/bin/env python3

"""
Read fastq files
create per-sample (sub)folder tree
run gene fusion predictions
start data collection
during all steps, perform slurm scheduling and process monitoring

@author: Tron (PASO), BNT (URLA)
"""
import os
import os.path
import time
from shutil import copy

import logzero
from logzero import logger

import easy_fuse
import easy_fuse.misc.io_methods as io_methods
from easy_fuse import __version__
from easy_fuse.misc import queueing
from easy_fuse.misc.config import EasyFuseConfiguration

DEFAULT_CPU_MEM = "1,16"

READ_FILTER_STEP = "readfilter"


class Processing(object):
    """Run, monitor and schedule fastq processing for fusion gene prediction"""

    def __init__(
        self,
        sample_id: str,
        fastq1: str,
        fastq2: str,
        working_dir,
        config: EasyFuseConfiguration,
        jobname_suffix,
    ):
        """Parameter initiation and work folder creation. Start of progress logging."""
        self.working_dir = os.path.join(os.path.abspath(working_dir), sample_id)
        io_methods.create_folder(self.working_dir)
        self.log_path = os.path.join(self.working_dir, "easyfuse_processing.log")
        logzero.logfile(self.log_path)

        logger.info("Starting easyfuse!")
        self.sample_id = sample_id
        assert os.path.exists(fastq1), "Fastq file {} does not exist!".format(fastq1)
        assert os.path.exists(fastq2), "Fastq file {} does not exist!".format(fastq2)
        self.fastq1 = os.path.abspath(fastq1)
        self.fastq2 = os.path.abspath(fastq2)

        self.cfg = config
        copy(self.cfg.config_file, working_dir)

        self.jobname_suffix = None
        if jobname_suffix:
            self.jobname_suffix = jobname_suffix

    # The run method simply greps and organises fastq input files.
    # Fastq pairs (single end input is currently not supported) are then send to "execute_pipeline"
    def run(self):
        """General parameter setting, identification of fastq files and initiation of processing"""
        logger.info("Pipeline Version: {}".format(__version__))

        ref_genome = self.cfg["general"]["ref_genome_build"]
        ref_trans = self.cfg["general"]["ref_trans_version"]

        logger.info(
            "Reference Genome: {0}, Reference Transcriptome: {1}".format(
                ref_genome, ref_trans
            )
        )

        self.execute_pipeline(self.fastq1, self.fastq2, self.sample_id, ref_genome, ref_trans)

        # summarize all data if selected
        if "summary" in self.cfg["general"]["tools"].split(","):
            dependency = []

            dependency.extend(
                queueing.get_jobs_by_name(
                    "requantifyBest-{}".format(self.sample_id),
                    self.cfg["general"]["queueing_system"],
                )
            )

            # TODO: add support for running without model and multiple models

            cmd_summarize = self._build_command_summarize_data()

            resources_summary = self.cfg["resources"]["summary"].split(",")
            cpu = resources_summary[0]
            mem = resources_summary[1]
            self.submit_job(
                "-".join(
                    [
                        self.cfg["general"]["pipeline_name"],
                        "Summary",
                        str(int(round(time.time()))),
                    ]
                ),
                cmd_summarize,
                cpu,
                mem,
                os.path.join(self.working_dir, "fusion_summary"),
                dependency,
                self.cfg["general"]["receiver"],
            )

    def _build_command_summarize_data(self):
        # summarize data input files
        context_seq_file = os.path.join(
            self.working_dir,
            "fetchdata",
            "fetched_contextseqs",
            "Context_Seqs.csv",
        )
        detect_fusion_file = os.path.join(
            self.working_dir,
            "fetchdata",
            "fetched_fusions",
            "Detected_Fusions.csv",
        )
        requant_mode = self.cfg["general"]["requant_mode"]
        requant_cpm_file = os.path.join(
            self.working_dir,
            "fetchdata",
            "classification",
            "classification_{}.tdt".format(requant_mode),
        )
        requant_cnt_file = os.path.join(
            self.working_dir,
            "fetchdata",
            "classification",
            "classification_{}.tdt.counts".format(requant_mode),
        )
        input_reads_stats_file = os.path.join(
            self.working_dir,
            "filtered_reads",
            "{}_Log.final.out".format(self.sample_id)
        )

        # summarize data output folder
        fusion_summary_output_folder = os.path.join(self.working_dir, "fusion_summary")
        io_methods.create_folder(fusion_summary_output_folder)
        cmd_summarize = (
            "easy-fuse summarize-data "
            "--input-fusions {input_fusions} "
            "--input-fusion-context-seqs {input_fusion_context_seqs} "
            "--input-requant-cpm {input_requant_cpm} "
            "--input-requant-counts {input_requant_counts} "
            "--input-reads-stats {input_reads_stats} "
            "--model_predictions "
            "--output-folder {output_folder} "
            "--requant-mode {requant_mode} "
            "--model-pred-threshold {model_pred_threshold} "
            "--fusion-tools {fusion_tools}".format(
                input_fusions=detect_fusion_file,
                input_fusion_context_seqs=context_seq_file,
                input_requant_cpm=requant_cpm_file,
                input_requant_counts=requant_cnt_file,
                input_reads_stats=input_reads_stats_file,
                output_folder=fusion_summary_output_folder,
                requant_mode=requant_mode,
                model_pred_threshold=self.cfg["general"]["model_pred_threshold"],
                fusion_tools=self.cfg["general"]["fusiontools"],
            )
        )
        return cmd_summarize

    # Per sample, define input parameters and execution commands, create a folder tree and submit runs to slurm
    def execute_pipeline(self, fq1, fq2, sample_id, ref_genome, ref_trans):
        """Create sample specific subfolder structure and run tools on fastq files"""

        refs = self.cfg["references"]

        # Genome/Gene references to use
        genome_fasta = refs["genome_fasta"]
        genome_chrs_path = refs["genome_fastadir"]
        genes_gtf_path = refs["genes_gtf"]
        genes_tsl_path = refs["genes_tsl"]
        genes_adb_path = refs["genes_adb"]

        # Path' to specific indices
        indices = self.cfg["indices"]
        other_files = self.cfg["other_files"]

        bowtie_index_path = indices["bowtie"]
        star_index_path = indices["star"]
        starfusion_index_path = indices["starfusion"]
        fusioncatcher_index_path = indices["fusioncatcher"]
        infusion_cfg_path = other_files["infusion_cfg"]

        # Output results folder creation - currently included:
        # 1) Gene/Isoform expression: star
        # 2) Fusion prediction: mapsplice, fusioncatcher, starfusion, infusion, soapfuse
        qc_path = os.path.join(self.working_dir, "qc")
        skewer_path = os.path.join(qc_path, "skewer")
        qc_table_path = os.path.join(qc_path, "qc_table.txt")
        filtered_reads_path = os.path.join(self.working_dir, "filtered_reads")
        expression_path = os.path.join(self.working_dir, "expression")
        star_path = os.path.join(expression_path, "star")
        fusion_path = os.path.join(self.working_dir, "fusion")
        mapsplice_path = os.path.join(fusion_path, "mapsplice")
        fusioncatcher_path = os.path.join(fusion_path, "fusioncatcher")
        starfusion_path = os.path.join(fusion_path, "starfusion")
        infusion_path = os.path.join(fusion_path, "infusion")
        soapfuse_path = os.path.join(fusion_path, "soapfuse")
        fetchdata_path = os.path.join(self.working_dir, "fetchdata")
        # former fetchdata.py
        detected_fusions_path = os.path.join(fetchdata_path, "fetched_fusions")
        detected_fusions_file = os.path.join(
            detected_fusions_path, "Detected_Fusions.csv"
        )
        context_seq_path = os.path.join(fetchdata_path, "fetched_contextseqs")
        context_seq_file = os.path.join(context_seq_path, "Context_Seqs.csv")
        context_seq_fasta = os.path.join(context_seq_path, "Context_Seqs.csv.fasta")
        star_genome_path = os.path.join(context_seq_path, "STAR_idx")
        star_align_path = os.path.join(context_seq_path, "STAR_align")
        star_align_file = os.path.join(star_align_path, "{}_".format(sample_id))
        classification_path = os.path.join(fetchdata_path, "classification")
        classification_file = os.path.join(classification_path, "classification")

        fastqc_1 = os.path.join(
            qc_path,
            os.path.basename(fq1).rstrip(".fastq.gz") + "_fastqc",
            "fastqc_data.txt",
        )
        fastqc_2 = os.path.join(
            qc_path,
            os.path.basename(fq2).rstrip(".fastq.gz") + "_fastqc",
            "fastqc_data.txt",
        )

        for folder in [
            qc_path,
            skewer_path,
            filtered_reads_path,
            expression_path,
            star_path,
            fusion_path,
            mapsplice_path,
            fusioncatcher_path,
            starfusion_path,
            infusion_path,
            soapfuse_path,
            fetchdata_path,
            detected_fusions_path,
            context_seq_path,
            star_genome_path,
            star_align_path,
            classification_path,
        ]:
            io_methods.create_folder(folder)

        # get a list of tools from the config file which shall be run on this sample
        tools = self.cfg["general"]["tools"].split(",")
        cmds = self.cfg["commands"]
        # Define cmd strings for each program
        # urla: mapsplice requires gunzip'd read files and process substitutions don't seem to work in slurm scripts...
        #       process substitution do somehow not work from this script - c/p the command line to the terminal, however, works w/o issues?!
        # (0) QC
        cmd_fastqc = "{0} --nogroup --extract -t 6 -o {1} {2} {3}".format(
            cmds["fastqc"], qc_path, fq1, fq2
        )
        cmd_qc_parser = (
            "easy-fuse qc-parser "
            "-i {0} {1} "
            "-o {2}".format(fastqc_1, fastqc_2, qc_table_path)
        )

        cmd_skewer = (
            "easy-fuse skewer-wrapper -q {0} -i {1} {2} -o {3} -b {4} -m {5}".format(
                qc_table_path,
                fq1,
                fq2,
                skewer_path,
                self.cfg["commands"]["skewer"],
                self.cfg["general"]["min_read_len_perc"],
            )
        )

        fq0 = ""  # TODO: change by None or get rid of
        if "qc" in tools:
            fq0 = os.path.join(skewer_path, "out_file-trimmed.fastq.gz")
            fq1 = os.path.join(skewer_path, "out_file-trimmed-pair1.fastq.gz")
            fq2 = os.path.join(skewer_path, "out_file-trimmed-pair2.fastq.gz")
        else:
            qc_table_path = "None"

        # (1) Readfilter
        cmd_star_filter = (
            "{0} "
            "--genomeDir {1} "
            "--outFileNamePrefix {2}_ "
            "--readFilesCommand zcat "
            "--readFilesIn {3} {4} "
            "--outFilterMultimapNmax 100 "
            "--outSAMmultNmax 1 "
            "--chimSegmentMin 10 "
            "--chimJunctionOverhangMin 10 "
            "--alignSJDBoverhangMin 10 "
            "--alignMatesGapMax {5} "
            "--alignIntronMax {5} "
            "--chimSegmentReadGapMax 3 "
            "--alignSJstitchMismatchNmax 5 -1 5 5 "
            "--seedSearchStartLmax 20 "
            "--winAnchorMultimapNmax 50 "
            "--outSAMtype BAM Unsorted "
            "--chimOutType Junctions WithinBAM "
            "--outSAMunmapped Within KeepPairs "
            "--runThreadN waiting_for_cpu_number".format(
                cmds["star"],
                star_index_path,
                os.path.join(filtered_reads_path, sample_id),
                fq1,
                fq2,
                self.cfg["general"]["max_dist_proper_pair"],
            )
        )

        cmd_read_filter = "easy-fuse read-filter --input {0}_Aligned.out.bam --output {0}_Aligned.out.filtered.bam".format(
            os.path.join(filtered_reads_path, sample_id)
        )

        # re-define fastq's if filtering is on (default)
        fq0 = ""
        if READ_FILTER_STEP in tools:
            fq0 = os.path.join(
                filtered_reads_path,
                os.path.basename(fq1)
                .replace("R1", "R0")
                .replace(".fastq.gz", "_filtered_singles.fastq.gz"),
            )
            fq1 = os.path.join(
                filtered_reads_path,
                os.path.basename(fq1).replace(".fastq.gz", "_filtered.fastq.gz"),
            )
            fq2 = os.path.join(
                filtered_reads_path,
                os.path.basename(fq2).replace(".fastq.gz", "_filtered.fastq.gz"),
            )

        cmd_bam_to_fastq = "{0} fastq -0 {1} -1 {2} -2 {3} --threads waiting_for_cpu_number {4}_Aligned.out.filtered.bam".format(
            cmds["samtools"],
            fq0,
            fq1,
            fq2,
            os.path.join(filtered_reads_path, sample_id),
        )

        # (2) Star expression quantification (required for starfusion)
        cmd_star = "{0} --genomeDir {1} --outFileNamePrefix waiting_for_output_string --runThreadN waiting_for_cpu_number --runMode alignReads --readFilesIn {2} {3} --readFilesCommand zcat --chimSegmentMin 10 --chimJunctionOverhangMin 10 --alignSJDBoverhangMin 10 --alignMatesGapMax {4} --alignIntronMax {4} --chimSegmentReadGapMax 3 --alignSJstitchMismatchNmax 5 -1 5 5 --seedSearchStartLmax 20 --winAnchorMultimapNmax 50 --outSAMtype BAM SortedByCoordinate --chimOutType Junctions SeparateSAMold --chimOutJunctionFormat 1".format(
            cmds["star"],
            star_index_path,
            fq1,
            fq2,
            self.cfg["general"]["max_dist_proper_pair"],
        )

        # (3) Mapslice
        # Using "gunzip -c" instead of "gunzip --keep" to keep compatibility with gunzip versions
        cmd_extr_fastq1 = "gunzip -c -f {0} > {1}".format(fq1, fq1[:-3])
        cmd_extr_fastq2 = "gunzip -c -f {0} > {1}".format(fq2, fq2[:-3])

        # Added python interpreter to circumvent external hardcoded shell script
        cmd_mapsplice = "{0} --chromosome-dir {1} -x {2} -1 {3} -2 {4} --threads waiting_for_cpu_number --output {5} --qual-scale phred33 --bam --seglen 20 --min-map-len 40 --gene-gtf {6} --fusion".format(
            cmds["mapsplice"],
            genome_chrs_path,
            bowtie_index_path,
            fq1[:-3],
            fq2[:-3],
            mapsplice_path,
            genes_gtf_path,
        )
        # (4) Fusioncatcher
        cmd_fusioncatcher = (
            "{0} --input {1} --data {2} --output {3} -p waiting_for_cpu_number".format(
                cmds["fusioncatcher"],
                ",".join([fq1, fq2]),
                fusioncatcher_index_path,
                fusioncatcher_path,
            )
        )

        # (5) Starfusion
        cmd_starfusion = "{0} --chimeric_junction {1} --genome_lib_dir {2} --CPU waiting_for_cpu_number --output_dir {3}".format(
            cmds["starfusion"],
            "{}_Chimeric.out.junction".format(os.path.join(star_path, sample_id)),
            starfusion_index_path,
            starfusion_path,
        )

        # (6) Infusion
        # TODO: Use these command line args for more sensitive results:
        # --min-split-reads 0 --min-fragments 0 --min-not-rescued 2
        cmd_infusion = (
            "{0} {1} -1 {2} -2 {3} --skip-finished --min-unique-alignment-rate 0 "
            "--min-unique-split-reads 0 --allow-non-coding --out-dir {4} {5}".format(
                cmds["python2"],
                cmds["infusion"],
                fq1,
                fq2,
                infusion_path,
                infusion_cfg_path,
            )
        )

        # (7) Soapfuse
        cmd_soapfuse = (
            "easy-fuse soapfuse-wrapper -q {0} -i {1} -o {2} -b {3} -c {4}".format(
                qc_table_path,
                " ".join([fq1, fq2]),
                soapfuse_path,
                self.cfg["commands"]["soapfuse"],
                self.cfg["other_files"]["soapfuse_cfg"],
            )
        )

        fusioncatcher_results_path_1 = os.path.join(
            self.working_dir, "fusion", "fusioncatcher", "summary_candidate_fusions.txt"
        )
        fusioncatcher_results_path_2 = os.path.join(
            self.working_dir, "fusion", "fusioncatcher", "final-list_candidate-fusion-genes.txt"
        )
        starfusion_results_path = os.path.join(
            self.working_dir, "fusion", "starfusion", "star-fusion.fusion_predictions.tsv"
        )
        mapsplice_results_path = os.path.join(
            self.working_dir, "fusion", "mapsplice", "fusions_well_annotated.txt"
        )
        infusion_results_path = os.path.join(
            self.working_dir, "fusion", "infusion", "fusions.detailed.txt"
        )
        soapfuse_results_path = os.path.join(
            self.working_dir, "fusion", "soapfuse", "final_fusion_genes", self.sample_id,
            "{}.final.Fusion.specific.for.genes".format(self.sample_id)
        )
        cmd_fusiondata = (
            "easy-fuse fusion-parser "
            "--input-fusioncatcher {input_fusion_catcher} "
            "--input-fusioncatcher2 {input_fusion_catcher2} "
            "--input-starfusion {input_star_fusion} "
            "--input-mapsplice {input_mapsplice} "
            "--input-infusion {input_infusion} "
            "--input-soapfuse {input_soapfuse} "
            "--output {output_folder} "
            "--sample {sample} "
            "--logger {logfile}".format(
                output_folder=detected_fusions_path,
                sample=sample_id,
                logfile=self.log_path,
                input_fusion_catcher=fusioncatcher_results_path_1,
                input_fusion_catcher2=fusioncatcher_results_path_2,
                input_star_fusion=starfusion_results_path,
                input_mapsplice=mapsplice_results_path,
                input_infusion=infusion_results_path,
                input_soapfuse=soapfuse_results_path
            )
        )

        cmd_contextseq = (
            "easy-fuse annotation "
            "--detected_fusions {0} "
            "--annotation_db {1} "
            "--out_csv {2} "
            "--genome_fasta {3} "
            "--tsl_info {4} "
            "--cis_near_dist {5} "
            "--context_seq_len {6} "
            "--tsl_filter_level {7}".format(
                detected_fusions_file,
                genes_adb_path,
                context_seq_file,
                genome_fasta,
                genes_tsl_path,
                self.cfg["general"]["cis_near_distance"],
                self.cfg["general"]["context_seq_len"],
                self.cfg["general"]["tsl_filter"],
            )
        )

        cmd_starindex = "easy-fuse star-index -i {0} -o {1} -t waiting_for_cpu_number -b {2}".format(
            context_seq_fasta, star_genome_path, cmds["star"]
        )

        cmd_staralign_fltr = (
            "{0} --genomeDir {1} --readFilesCommand zcat --readFilesIn {2} {3} "
            "--outSAMtype BAM SortedByCoordinate --outFilterMultimapNmax -1 "
            "--outSAMattributes Standard --outSAMunmapped None "
            "--outFilterMismatchNoverLmax 0.02 --runThreadN waiting_for_cpu_number "
            "--outFileNamePrefix {4}fltr_ --limitBAMsortRAM 48000000000".format(
                cmds["star"], star_genome_path, fq1, fq2, star_align_file
            )
        )

        cmd_bamindex_fltr = "{0} index {1}fltr_Aligned.sortedByCoord.out.bam".format(
            cmds["samtools"], star_align_file
        )

        input_reads_stats_file = os.path.join(
            filtered_reads_path,
            "{}_Log.final.out".format(sample_id)
        )
        cmd_requantify_fltr = (
            "easy-fuse requantify "
            "-i {0}fltr_Aligned.sortedByCoord.out.bam "
            "-o {1}_fltr.tdt "
            "-d 10 "
            "--input-reads-stats {2}".format(star_align_file, classification_file, input_reads_stats_file)
        )

        cmd_staralign_org = (
            "{0} --genomeDir {1} --readFilesCommand zcat --readFilesIn {2} {3} "
            "--outSAMtype BAM SortedByCoordinate --outFilterMultimapNmax -1 "
            "--outSAMattributes Standard --outSAMunmapped None --outFilterMismatchNoverLmax 0.02 "
            "--runThreadN waiting_for_cpu_number --outFileNamePrefix {4}org_ "
            "--limitBAMsortRAM 48000000000".format(
                cmds["star"], star_genome_path, fq1, fq2, star_align_file
            )
        )

        cmd_bamindex_org = "{0} index {1}org_Aligned.sortedByCoord.out.bam".format(
            cmds["samtools"], star_align_file
        )

        cmd_requantify_org = (
            "easy-fuse requantify "
            "-i {0}org_Aligned.sortedByCoord.out.bam "
            "-o {1}_org.tdt "
            "-d 10 "
            "--input-reads-stats {2}".format(star_align_file, classification_file, input_reads_stats_file)
        )

        # for testing, based on debug. should be removed if merged to original
        cmd_read_filter2 = (
            "easy-fuse requantify-filter "
            "--input {0}_Aligned.out.bam "
            "--input2 {1}.debug "
            "--input-reads-stats {2} "
            "--output {0}_Aligned.out.filtered2.bam".format(
                os.path.join(filtered_reads_path, sample_id),
                context_seq_file,
                input_reads_stats_file
            )
        )

        # re-define fastq's if filtering is on (default)
        fq0 = ""
        if "readfilter" in tools:
            fq0 = os.path.join(
                filtered_reads_path,
                os.path.basename(fq1)
                .replace("R1", "R0")
                .replace(".fastq.gz", "_filtered2_singles.fastq.gz"),
            )
            fq1 = os.path.join(
                filtered_reads_path,
                os.path.basename(fq1).replace(".fastq.gz", "_filtered2.fastq.gz"),
            )
            fq2 = os.path.join(
                filtered_reads_path,
                os.path.basename(fq2).replace(".fastq.gz", "_filtered2.fastq.gz"),
            )
        cmd_bam_to_fastq_fd = "{0} fastq -0 {1} -1 {2} -2 {3} --threads waiting_for_cpu_number {4}_Aligned.out.filtered2.bam".format(
            cmds["samtools"],
            fq0,
            fq1,
            fq2,
            os.path.join(filtered_reads_path, sample_id),
        )

        # allow soft-clipping? Specificity? --alignEndsType EndToEnd
        cmd_staralign_best = (
            "{0} "
            "--genomeDir {1} "
            "--readFilesCommand zcat "
            "--readFilesIn {2} {3} "
            "--outSAMtype BAM SortedByCoordinate "
            "--outFilterMultimapNmax -1 "
            "--outSAMattributes Standard "
            "--outSAMunmapped None "
            "--outFilterMismatchNoverLmax 0.02 "
            "--runThreadN waiting_for_cpu_number "
            "--outFileNamePrefix {4}best_ "
            "--limitBAMsortRAM 48000000000".format(
                cmds["star"], star_genome_path, fq1, fq2, star_align_file
            )
        )

        cmd_bamindex_best = "{0} index " "{1}best_Aligned.sortedByCoord.out.bam".format(
            cmds["samtools"], star_align_file
        )

        cmd_requantify_best = (
            "easy-fuse requantify "
            "-i {0}best_Aligned.sortedByCoord.out.bam "
            "-o {1}_best.tdt "
            "-d 10 "
            "--input-reads-stats {2}".format(star_align_file, classification_file, input_reads_stats_file)
        )

        # set final lists of executable tools and path
        exe_tools = [
            "qc",  # 0
            READ_FILTER_STEP,  # 1
            "star",
            "mapsplice",
            "fusioncatcher",
            "starfusion",
            "infusion",
            "soapfuse",
            "fusiongrep",
            "contextseq",
            "starindex",
            "staralignFltr",
            "bamindexFltr",
            "requantifyFltr",
            "staralignOrg",
            "bamindexOrg",
            "requantifyOrg",
            "readFilter2",
            "readFilter2b",
            "staralignBest",
            "bamindexBest",
            "requantifyBest",
        ]
        exe_cmds = [
            " && ".join([cmd_fastqc, cmd_qc_parser, cmd_skewer]),
            " && ".join([cmd_star_filter, cmd_read_filter, cmd_bam_to_fastq]),
            cmd_star,
            " && ".join([cmd_extr_fastq1, cmd_extr_fastq2, cmd_mapsplice]),
            cmd_fusioncatcher,
            cmd_starfusion,
            cmd_infusion,
            cmd_soapfuse,
            cmd_fusiondata,
            cmd_contextseq,
            cmd_starindex,
            cmd_staralign_fltr,
            cmd_bamindex_fltr,
            cmd_requantify_fltr,
            cmd_staralign_org,
            cmd_bamindex_org,
            cmd_requantify_org,
            cmd_read_filter2,
            cmd_bam_to_fastq_fd,
            cmd_staralign_best,
            cmd_bamindex_best,
            cmd_requantify_best,
        ]
        exe_path = [
            qc_path,
            filtered_reads_path,
            star_path,
            mapsplice_path,
            fusioncatcher_path,
            starfusion_path,
            infusion_path,
            soapfuse_path,
            classification_path,
            detected_fusions_path,
            context_seq_path,
            star_genome_path,
            star_align_path,
            star_align_path,
            classification_path,
            star_align_path,
            star_align_path,
            classification_path,
            filtered_reads_path,
            filtered_reads_path,
            star_align_path,
            star_align_path,
            classification_path,
        ]

        # create and submit slurm job if the tool is requested and hasn't been run before
        for i, tool in enumerate(exe_tools, 0):
            if tool in tools:
                # check dependencies of the pipeline.
                # Besides tool dependencies (Starfusion -> Star), read filtering is mandatory
                if tool == READ_FILTER_STEP and READ_FILTER_STEP not in tools:
                    logger.error(
                        """Error 99: Sample {} will be skipped due to missing read filtering.\n
                            Read filtering is currently a mandatory step for the processing.\n
                            Because you haven't run it before for this sample, you have to include \"Readfilter\" in the tool selection in your config.\n
                            """.format(
                            sample_id
                        )
                    )
                    logger.error(
                        "Error 99: Sample {} will be skipped due to missing read filtering.".format(
                            sample_id
                        )
                    )
                    return 0
                elif tool == "starfusion" and "star" not in tools:
                    logger.error(
                        """Error 99: Running {0} for Sample {1} will be skipped due to a missing dependency.\n
                            {0} builds on Star and it is therefore mandatory to run this first.\n
                            Because you haven't run it before for this sample, you have to include \"Star\" in the tool selection in your config.\n
                            """.format(
                            sample_id
                        )
                    )
                    logger.error(
                        "Error 99: Running {0} for Sample {1} will be skipped due to a missing dependency.".format(
                            tool, sample_id
                        )
                    )
                    continue

                # prepare slurm jobs: get ressources, create uid, set output path and check dependencies
                logger.debug("Submitting {} run".format(tool))
                cpumem = (
                    self.cfg["resources"].get(tool.lower(), DEFAULT_CPU_MEM).split(",")
                )
                cpu = cpumem[0]
                mem = cpumem[1]
                uid = "-".join([self.cfg["general"]["pipeline_name"], tool, sample_id])
                if tool == "star":
                    exe_cmds[i] = (
                        exe_cmds[i]
                        .replace(
                            "waiting_for_output_string",
                            "{}_".format(os.path.join(exe_path[i], sample_id)),
                        )
                        .replace("waiting_for_cpu_number", str(cpu))
                    )
                else:
                    exe_cmds[i] = (
                        exe_cmds[i]
                        .replace("waiting_for_output_string", exe_path[i])
                        .replace("waiting_for_cpu_number", str(cpu))
                    )
                cmd = exe_cmds[i]

                # Build slurm dependencies
                dependencies = self.build_dependencies(tool, sample_id)
                self.submit_job(uid, cmd, cpu, mem, exe_path[i], dependencies, "")
            else:
                logger.info(
                    "Skipping {0} as it is not selected for execution (Selected are: {1})".format(
                        tool, tools
                    )
                )

    def submit_job(
        self, uid, cmd, cores, mem_usage, output_results_folder, dependencies, mail
    ):
        """Submit job to for process generation"""
        logger.debug(
            "Submitting job: CMD - {0}; PATH - {1}; DEPS - {2}".format(
                cmd, output_results_folder, dependencies
            )
        )
        que_sys = self.cfg["general"]["queueing_system"]
        already_running = queueing.get_jobs_by_name(uid, que_sys)
        if not already_running:
            # urla: for compatibility reasons (and to be independent of shell commands), concatenated commands are splitted again,
            #       dependencies within the splitted groups updated and everything submitted sequentially to the queueing system
            module_file = self.cfg["general"]["build_env"]

            for i, cmd_split in enumerate(cmd.split(" && ")):
                if not que_sys in ["slurm", "pbs"]:
                    cmd_split = cmd_split.split(" ")
                dependencies.extend(
                    queueing.get_jobs_by_name("{0}_CMD{1}".format(uid, i - 1), que_sys)
                )
                if self.jobname_suffix:
                    queueing.submit(
                        "{0}_CMD{1}-{2}".format(uid, i, self.jobname_suffix),
                        cmd_split,
                        cores,
                        mem_usage,
                        output_results_folder,
                        dependencies,
                        self.cfg["general"]["partition"],
                        self.cfg["general"]["user"],
                        self.cfg["general"]["time_limit"],
                        mail,
                        module_file,
                        que_sys,
                    )
                else:
                    queueing.submit(
                        "{0}_CMD{1}".format(uid, i),
                        cmd_split,
                        cores,
                        mem_usage,
                        output_results_folder,
                        dependencies,
                        self.cfg["general"]["partition"],
                        self.cfg["general"]["user"],
                        self.cfg["general"]["time_limit"],
                        mail,
                        module_file,
                        que_sys,
                    )
                time.sleep(0.5)
        else:
            logger.error(
                "A job with this application/sample combination is currently running. Skipping {} in order to avoid unintended data loss.".format(
                    uid
                )
            )

    def build_dependencies(self, tool, sample_id):
        slurm_dep = {
            "qc": [],
            "readfilter": ["qc"],
            "star": ["readfilter"],
            "starfusion": ["star"],
            "mapsplice": ["readfilter"],
            "fusioncatcher": ["readfilter"],
            "infusion": ["readfilter"],
            "soapfuse": ["readfilter"],
            "fusiongrep": [
                "starfusion",
                "mapsplice",
                "fusioncatcher",
                "infusion",
                "soapfuse",
            ],
            "contextseq": ["fusiongrep"],
            "starindex": ["contextseq"],
            "readFilter2": ["contextseq"],
            "readFilter2b": ["readFilter2"],
            "staralignBest": ["starindex", "readFilter2b"],
            "bamindexBest": ["staralignBest"],
            "requantifyBest": ["bamindexBest"],
        }
        dependencies = []
        que_sys = self.cfg["general"]["queueing_system"]

        for pre_tool in slurm_dep[tool]:
            dependencies.extend(
                queueing.get_jobs_by_name("{}-{}".format(pre_tool, sample_id), que_sys)
            )

        return dependencies


def add_pipeline_parser_args(pipeline_parser):
    pipeline_parser.add_argument(
        "-1",
        "--fastq1",
        dest="fastq1",
        help="FASTQ file for read 1",
        required=True,
    )
    pipeline_parser.add_argument(
        "-2",
        "--fastq2",
        dest="fastq2",
        help="FASTQ file for read 2",
        required=True,
    )
    pipeline_parser.add_argument(
        "-s",
        "--sample-id",
        dest="sample_id",
        help="Sample identifier",
        required=True,
    )
    pipeline_parser.add_argument(
        "-o",
        "--output-folder",
        dest="output_folder",
        help="Specify full path of the folder to save the results into.",
        required=True,
    )
    # optional arguments
    pipeline_parser.add_argument(
        "-c",
        "--config-file",
        dest="config_file",
        required=True,
        help="Specify alternative config file to use for your analysis",
    )
    pipeline_parser.add_argument(
        "-p",
        "--jobname_suffix",
        dest="jobname_suffix",
        help="Specify a jobname suffix for running jobs on a queueing system",
        default="",
    )
    pipeline_parser.add_argument(
        "-V",
        "--version",
        dest="version",
        action="version",
        version="%(prog)s {version}".format(version=easy_fuse.__version__),
        help="Get version info",
    )
    pipeline_parser.set_defaults(func=pipeline_command)


def pipeline_command(args):
    if args.jobname_suffix:
        "-p {}".format(args.jobname_suffix)

    config = EasyFuseConfiguration(args.config_file)
    Processing(
        args.sample_id, args.fastq1, args.fastq2, args.output_folder, config, args.jobname_suffix
    ).run()



import unittest

from easy_fuse.fusionannotation import FusionAnnotation


class TestFusionAnnotation(unittest.TestCase):
    def test_define_type(self):
        self.assertEqual(
            FusionAnnotation.define_type(500, "1", 200, "+", "2", 600, "-"), "trans_inv"
        )
        self.assertEqual(
            FusionAnnotation.define_type(500, "1", 200, "+", "1", 800, "+"), "cis_far"
        )
        self.assertEqual(
            FusionAnnotation.define_type(1000, "1", 200, "+", "1", 800, "+"), "cis_near"
        )
        self.assertEqual(
            FusionAnnotation.define_type(500, "4", 800, "+", "4", 200, "+"), "cis_trans"
        )
        self.assertEqual(
            FusionAnnotation.define_type(500, "7", 800, "-", "7", 200, "-"), "cis_far"
        )
        self.assertEqual(
            FusionAnnotation.define_type(500, "2", 800, "-", "3", 200, "-"), "trans"
        )

    def test_check_exon_overlap(self):
        wt1_exon_pos_list = [(1, 100), (200, 300)]
        wt2_exon_pos_list = [(500, 600), (700, 800)]
        self.assertEqual(
            FusionAnnotation.check_exon_overlap(
                wt1_exon_pos_list, wt2_exon_pos_list, "cis_near"
            ),
            False,
        )

        wt1_exon_pos_list = [(1, 100), (200, 300)]
        wt2_exon_pos_list = [(150, 600), (700, 800)]
        self.assertEqual(
            FusionAnnotation.check_exon_overlap(
                wt1_exon_pos_list, wt2_exon_pos_list, "cis_near"
            ),
            True,
        )

        wt1_exon_pos_list = [(50, 100), (200, 300)]
        wt2_exon_pos_list = [(301, 600), (700, 800)]
        self.assertEqual(
            FusionAnnotation.check_exon_overlap(
                wt1_exon_pos_list, wt2_exon_pos_list, "cis_far"
            ),
            False,
        )

        wt1_exon_pos_list = [(400, 600), (800, 1000)]
        wt2_exon_pos_list = [(200, 300), (400, 500)]
        self.assertEqual(
            FusionAnnotation.check_exon_overlap(
                wt1_exon_pos_list, wt2_exon_pos_list, "cis_trans"
            ),
            True,
        )

        wt1_exon_pos_list = [(1, 100), (200, 300)]
        wt2_exon_pos_list = [(500, 600), (700, 800)]
        self.assertEqual(
            FusionAnnotation.check_exon_overlap(
                wt1_exon_pos_list, wt2_exon_pos_list, "trans"
            ),
            False,
        )


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3

import os
import subprocess
import sys
from shutil import copyfile
from logzero import logger


def run_skewer(skewer_bin, min_rl_perc, qc_table, input, output):

    remaining_len = 1000
    read_len = 0
    with open(qc_table) as inf:
        next(inf)
        for line in inf:
            elements = line.rstrip().split(",")
            read_len = int(elements[1])
            remaining = int(elements[3])
            if remaining < remaining_len:
                remaining_len = remaining

    if remaining_len != read_len and remaining_len >= (min_rl_perc * read_len):
        cmd = "{} -l {} -q 28 -m pe -t 6 -k 5 -z -o {} {}".format(
            skewer_bin,
            remaining_len,
            os.path.join(output, "out_file"),
            " ".join(input),
        )

        proc = subprocess.Popen(
            cmd, shell=True, stderr=subprocess.PIPE, stdout=subprocess.PIPE
        )
        (_, stderrdata) = proc.communicate()
        r = proc.returncode
        if r != 0:
            logger.error("Error within skewer")
            logger.error(stderrdata)
            sys.exit(1)

    elif remaining_len < min_rl_perc * read_len:
        logger.error("Trim length too long. Discarding fastqs ({}).".format(input))
        open(os.path.join(output, "to_be_discarded"), "a").close()
        sys.exit(1)
    elif remaining_len == read_len:
        if len(input) == 2:
            out_1 = os.path.join(output, "out_file-trimmed-pair1.fastq.gz")
            out_2 = os.path.join(output, "out_file-trimmed-pair2.fastq.gz")
            logger.info("Nothing to trim. Copying FASTQs")
            logger.info("cp {} {}".format(input[0], out_1))
            logger.info("cp {} {}".format(input[1], out_2))
            try:
                copyfile(input[0], out_1)
            #                os.symlink(args.input[0], out_1)
            except OSError:
                logger.error("Copy of {} already exists".format(out_1))
            try:
                copyfile(input[1], out_2)
            #                os.symlink(args.input[1], out_2)
            except OSError:
                logger.error("Copy of {} already exists".format(out_2))


def add_skewer_args(parser):
    parser.add_argument(
        "-q",
        "--qc-table",
        dest="qc_table",
        help="Specify input QC table",
        required=True,
    )
    parser.add_argument(
        "-i",
        "--input",
        dest="input",
        nargs="+",
        help="Specify input FASTQ files",
        required=True,
    )
    parser.add_argument(
        "-o", "--output", dest="output", help="Specify output folder", default="."
    )
    parser.add_argument(
        "-b", "--binary", dest="binary", help="Specify Skewer binary", required=True
    )
    parser.add_argument(
        "-m",
        "--min-read-len-perc",
        dest="min_read_len_perc",
        type=float,
        help="Specify minimum read length percentage",
        default=0.75,
    )
    parser.set_defaults(func=skewer_command)


def skewer_command(args):
    run_skewer(
        skewer_bin=args.binary,
        min_rl_perc=args.min_read_len_perc,
        qc_table=args.qc_table,
        input=args.input,
        output=args.output,
    )



import os
from configparser import ConfigParser


class EasyFuseConfiguration(ConfigParser):
    def __init__(self, config_file):
        super().__init__()
        self.config_file = os.path.abspath(config_file)
        assert os.path.exists(
            self.config_file
        ), "Config file {} does not exist!".format(self.config_file)
        self.load_configuration()

    def load_configuration(self):
        self.read(self.config_file)



#!/usr/bin/env python

import gzip
import os
import subprocess
import sys

from logzero import logger


def run_soapfuse(binary, config_file, qc_table, input, output):

    cols = input[0].rsplit("/", 3)
    remaining_len = 1000
    if qc_table != "None":

        with open(qc_table) as inf:
            inf.readline()
            for line in inf:
                elements = line.rstrip().split(",")
                remaining = int(elements[3])
                if remaining < remaining_len:
                    remaining_len = remaining
    else:
        with gzip.open(input[0]) as inf:
            next(inf)
            line = inf.readline()
            remaining_len = len(line.rstrip())

    cfg_file = os.path.join(output, "samples.csv")
    cfg_out = open(cfg_file, "w")

    cfg_out.write("{}\t{}\tS\t{}\n".format(cols[1], cols[2], remaining_len))
    cfg_out.close()
    for i, file in enumerate(input):
        try:
            os.symlink(
                file,
                os.path.join(
                    cols[0], cols[1], cols[2], "S_{}.fastq.gz".format((i + 1))
                ),
            )
        except:
            logger.info("Symlink already exists!")
    cmd = "{} -c {} -fd {} -l {} -o {}".format(
        binary, config_file, cols[0], cfg_file, output
    )

    proc = subprocess.Popen(
        cmd, shell=True, stderr=subprocess.PIPE, stdout=subprocess.PIPE
    )
    (_, stderrdata) = proc.communicate()
    r = proc.returncode
    if r != 0:
        logger.error(stderrdata)
        sys.exit(1)


def add_soapfuse_wrapper_args(parser):
    parser.add_argument(
        "-q", "--qc-table", dest="qc_table", help="Specify input QC table"
    )
    parser.add_argument(
        "-i",
        "--input",
        dest="input",
        nargs="+",
        help="Specify input FASTQ files",
        required=True,
    )
    parser.add_argument(
        "-o", "--output", dest="output", help="Specify output folder", default="."
    )
    parser.add_argument(
        "-b", "--binary", dest="binary", help="Specify Soapfuse binary", required=True
    )
    parser.add_argument(
        "-c",
        "--config-file",
        dest="config_file",
        help="Specify Soapfuse config file to use for your analysis",
        required=True,
    )
    parser.set_defaults(func=soapfuse_wrapper_command)


def soapfuse_wrapper_command(args):
    run_soapfuse(
        binary=args.binary,
        config_file=args.config_file,
        qc_table=args.qc_table,
        input=args.input,
        output=args.output,
    )



#!/usr/bin/env python

"""
Create (if necessary) a sqlite3 database from an ensembl gff3 file.
Afterwards, test whether the db can be accessed and print the counts
of all features in it

@author: BNT (URLA)
@version: 20190704
"""

from __future__ import print_function, division

import os

import gffutils
from logzero import logger


# pylint: disable=line-too-long
#         yes they are partially, but I do not consider this to be relevant here
class GTF2DB(object):
    """stub"""

    def __init__(self, gff_in_file, db_out_file):
        """Parameter initialization"""
        self.gff_in_file = gff_in_file
        self.db_out_file = db_out_file

    def create_gffdb(self, overwrite):
        """Create a gffutils database from an ensembl gff3 file"""
        if os.path.exists(self.db_out_file):
            if overwrite:
                logger.debug(
                    "File {} already exists, but overwrite is specified. Trying to delete the existing file".format(
                        self.db_out_file
                    )
                )
                os.remove(self.db_out_file)
                self.create_gffdb(overwrite)
            else:
                logger.debug(
                    "File {} already exists and overwrite is NOT specified.".format(
                        self.db_out_file
                    )
                )
        else:
            logger.debug(
                "File {} does not exist. Trying to create it from the provided gff file (This may take 10-20+ minutes)...".format(
                    self.db_out_file
                )
            )
            # this is a one-time operation and tailored to human ensembl gff3
            gffutils.create_db(
                self.gff_in_file,
                self.db_out_file,
                force_gff=True,
                force=True,
                merge_strategy="create_unique",
            )

        logger.info("Testing whether db can be accessed...")
        self.test_db()

    def test_db(self):
        """Test whether an existing database can be loaded and accessed"""
        gff3_db = gffutils.FeatureDB(self.db_out_file)
        for feature in gff3_db.featuretypes():
            logger.info(
                "Found {} features of type {} in the db.".format(
                    gff3_db.count_features_of_type(feature), feature
                )
            )


def add_gff3_to_db_args(parser):
    """Add gff3_to_db arguments to parser"""
    parser.add_argument(
        "--gff_input", dest="gff_in", help="ensembl gff3 annotation file", required=True
    )
    parser.add_argument(
        "--db_output",
        dest="db_out",
        help="gff3 derived annotation database from gffutils",
        required=True,
    )
    parser.add_argument(
        "--overwrite",
        dest="overwrite",
        help="Overwrite existing database if present",
        default=False,
        action="store_true",
    )
    parser.set_defaults(func=gff3_to_db_command)


def gff3_to_db_command(args):
    """Run gff3_to_db conversion"""
    GTF2DB(args.gff_in, args.db_out).create_gffdb(args.overwrite)





#!/usr/bin/env python3

"""
Get predicted gene fusions from tools
Fusions are parsed into "id / fusion genes / breakpoints / supporting reads / prediction tool"
and output based on their set recurrency value. The class requires the path' of data and
output folders and writes the "Detected_Fusions.csv"

@author: Tron (PASO)
@version: 20221109
"""
import re
import operator


# TODO: this fixes the reference genome! -> make it more flexible
chr_list = (
    "1",
    "2",
    "3",
    "4",
    "5",
    "6",
    "7",
    "8",
    "9",
    "10",
    "11",
    "12",
    "13",
    "14",
    "15",
    "16",
    "17",
    "18",
    "19",
    "20",
    "21",
    "22",
    "X",
    "Y",
    "MT",
)


def parse_fusioncatcher_results(infile1, infile2):
    """Load and parse results from fusioncatcher"""

    reciprocal_fusions = []
    with open(infile1) as predict_summary:
        for line in predict_summary:
            if line.strip().startswith("*"):
                fusion_gene = re.search(r"\*\s([\S]*)", line).group(1)
                if "reciprocal" in line:
                    reciprocal_fusions.append(fusion_gene.replace("--", "_").upper())
    fusion_map = {}
    with open(infile2) as prediction:
        next(prediction)  # skip header line
        for line in prediction:
            elements = line.rstrip().split("\t")

            fusion_gene = (elements[0] + "_" + elements[1]).upper()
            # if the fusion gene is reciprocal, the fusion id is reversed? <- what for??
            if fusion_gene in reciprocal_fusions:
                fusion_gene = (elements[1] + "_" + elements[0]).upper()

            # skip all prediction not on standard chromosomes
            if (
                elements[8].split(":")[0] not in chr_list
                or elements[9].split(":")[0] not in chr_list
            ):
                continue

            bpid = elements[8] + "_" + elements[9]

            fusion_map[bpid] = [
                fusion_gene,  # fusion_gene
                elements[8],  # up_gene_bp
                elements[9],  # dn_gene_bp
                elements[5],  # junc_reads_num
                elements[4],  # span_reads_num
            ]
    return fusion_map


def parse_starfusion_results(infile):
    """Load and parse results from star-fusion"""
    fusion_map = {}
    with open(infile, "r") as prediction:
        next(prediction)  # skip header line
        for line in prediction:
            elements = line.rstrip().split("\t")

            fusion_gene = elements[0].replace("--", "_").upper()
            # check whether fusion gene is not on primary chr
            if (
                elements[5].split(":")[0] not in chr_list
                or elements[7].split(":")[0] not in chr_list
            ):
                continue

            bpid = elements[5] + "_" + elements[7]

            fusion_map[bpid] = [
                fusion_gene,  # fusion_gene
                elements[5],  # up_gene_bp
                elements[7],  # dn_gene_bp
                elements[1],  # junc_reads_num
                elements[2],  # span_reads_num
            ]
    return fusion_map


def parse_mapsplice_results(infile):
    """Load and parse results from mapsplice"""
    fusion_map = {}
    with open(infile) as prediction:
        for line in prediction:
            elements = line.rstrip().split("\t")

            fusion_gene = (
                elements[60].split(",")[0] + "_" + elements[61].split(",")[0]
            ).upper()

            # element[0] = chr num, [1/2] = breakpoints, [5] = strands
            up_gene_id = (
                elements[0].split("~")[0] + ":" + elements[1] + ":" + elements[5][0]
            )
            dn_gene_id = (
                elements[0].split("~")[1] + ":" + elements[2] + ":" + elements[5][1]
            )

            if (
                up_gene_id.split(":")[0] not in chr_list
                or dn_gene_id.split(":")[0] not in chr_list
            ):
                continue

            bpid = up_gene_id + "_" + dn_gene_id

            fusion_map[bpid] = [
                fusion_gene,  # fusion_gene
                up_gene_id,  # up_gene_bp
                dn_gene_id,  # dn_gene_bp
                elements[4],  # junc_reads_num
                elements[27],  # span_reads_num
            ]
    return fusion_map


def parse_infusion_results(infile):
    """Load and parse results from infusion"""
    fusion_map = {}
    with open(infile) as prediction:
        next(prediction)  # skip header line
        for line in prediction:
            elements = line.rstrip().split("\t")

            fusion_gene = "{0}_{1}".format(
                get_fusion_gene_id_infusion(elements[21], elements[22]),
                get_fusion_gene_id_infusion(elements[27], elements[28]),
            ).upper()

            # element[1/4] = chr num, [2/5] = breakpoints, [23/29] = strands
            up_gene_id = elements[1] + ":" + elements[2] + ":" + elements[23]
            dn_gene_id = elements[4] + ":" + elements[5] + ":" + elements[29]

            # check whether fusion gene is not on primary chr
            if elements[1] not in chr_list or elements[4] not in chr_list:
                continue

            bpid = up_gene_id + "_" + dn_gene_id

            fusion_map[bpid] = [
                fusion_gene,  # fusion_gene
                up_gene_id,  # up_gene_bp
                dn_gene_id,  # dn_gene_bp
                elements[7],  # junc_reads_num
                elements[8],  # span_reads_num
            ]
    return fusion_map


def get_fusion_gene_id_infusion(gene_id, transcript_list_field):
    """Helper method for infusion data parsing. Returns the most frequently listed fusion id in a list of fusion ids"""
    # if only 1 gene id is returned as fusion partner, return this
    if not ";" in gene_id:
        return gene_id.replace("_", "-")
    # if 2 or more gene ids are given, take the most frequent from the transcript list
    gid_dict = {}
    for gid in transcript_list_field.split(";"):
        id_test = gid.split(":")[0][:-4]
        if not id_test in gid_dict:
            gid_dict[id_test] = 0
        gid_dict[id_test] += 1

    best_hit = max(iter(gid_dict.items()), key=operator.itemgetter(1))[0]

    return best_hit.replace("_", "-")


def parse_soapfuse_results(infile):
    """Load and parse results from soapfuse"""

    fusion_map = {}
    with open(infile) as prediction:
        next(prediction)  # skip header line
        for line in prediction:
            elements = line.rstrip().split("\t")

            fusion_gene = (elements[0] + "_" + elements[5]).upper()

            # element[1/6] = chr num, [3/8] = breakpoints, [2/7] = strands
            up_gene_id = elements[1] + ":" + elements[3] + ":" + elements[2]
            dn_gene_id = elements[6] + ":" + elements[8] + ":" + elements[7]

            # check whether fusion gene is not on primary chr
            if elements[1] not in chr_list or elements[6] not in chr_list:
                continue

            bpid = up_gene_id + "_" + dn_gene_id

            fusion_map[bpid] = [
                fusion_gene,  # fusion_gene
                up_gene_id,  # up_gene_bp
                dn_gene_id,  # dn_gene_bp
                elements[11],  # junc_reads_num
                elements[10],  # span_reads_num
            ]
    return fusion_map



#!/usr/bin/env python

"""
Get Transcript Support Level (TSL) info from an ensembl GTF file

@author: BNT (URLA)
@version: 20190702
"""

import sys

from logzero import logger


class Gtf2Tsl(object):
    """stub"""

    def __init__(self, gtf_in, tsl_out):
        """Parameter initialization"""
        self.gtf_in = gtf_in
        self.tsl_out = tsl_out

    def get_data(self):
        """stub"""
        feature_list = []
        tsl_dict = {}
        count_transcripts = 0
        with open(self.gtf_in, "r") as in_file:
            for line in in_file:
                if line.startswith("#"):  # skip header
                    continue
                _, _, feature, _, _, _, _, _, attribute = line.split("\t")
                if not feature in feature_list:
                    feature_list.append(feature)
                if feature == "transcript":
                    count_transcripts += 1
                    field_split = attribute.split(";")
                    transcript_id = "NA"
                    trans_biotype = "NA"
                    gene_biotype = "NA"
                    tsl = "NA"
                    for field in field_split:
                        transcript_id = self.get_field_value(
                            transcript_id, field, "transcript_id"
                        )
                        trans_biotype = self.get_field_value(
                            trans_biotype, field, "transcript_biotype"
                        )
                        gene_biotype = self.get_field_value(
                            gene_biotype, field, "gene_biotype"
                        )
                        tsl = self.get_field_value(
                            tsl, field, "transcript_support_level"
                        )

                    logger.info(
                        "{}_{}_{}_{}".format(
                            transcript_id, trans_biotype, gene_biotype, tsl
                        )
                    )
                    # check again, that what we have is what we expect
                    if not (
                        transcript_id[0:3] == "ENS" or transcript_id[0:4] == "MGP_"
                    ):
                        logger.error(
                            "This does not look like a valid ensembl id: {}. Is this a valid ensembl gtf line: {}?".format(
                                transcript_id, line
                            )
                        )
                        sys.exit(99)
                    if not transcript_id in tsl_dict:
                        tsl_dict[transcript_id] = [trans_biotype, gene_biotype, tsl]
                    if count_transcripts == 20:
                        break

        # write data to file
        with open(self.tsl_out, "w") as out_file:
            out_file.write(
                "{}\n".format(
                    "\t".join(["transcript_id", "trans_biotype", "gene_biotype", "tsl"])
                )
            )
            for transcript_id in tsl_dict:
                out_file.write(
                    "{}\t{}\n".format(transcript_id, "\t".join(tsl_dict[transcript_id]))
                )
        # list of available feature strings
        logger.info(feature_list)

    def get_field_value(self, old_value, field_string, lookup_string):
        """Helper method to get the value from a key-value pair in the form of: ( this the value "and this the key")"""
        if not old_value == "NA":  # check if value has been found previously
            return old_value
        new_value = "NA"
        field_string = field_string.strip()  # remove leading/trailing whitespaces
        if field_string.startswith(lookup_string):
            new_value = field_string.split('"')[
                1
            ]  # get the value inside quotation marks
        return new_value


def add_gtf_to_tsl_args(parser):
    """Add gtf_to_tsl arguments to parser"""
    parser.add_argument("--gtf", dest="gtf", help="Gtf input file", required=True)
    parser.add_argument("--tsl", dest="tsl", help="Tsl output file", required=True)
    parser.set_defaults(func=gtf_to_tsl_command)


def gtf_to_tsl_command(args):
    """Run gtf_to_tsl conversion"""
    Gtf2Tsl(args.gtf, args.tsl).get_data()



#!/usr/bin/env python3


class Parser(object):
    def __init__(self):
        pass

    def parse_total_sequences(self, infile):
        total_sequences = 0
        with open(infile) as inf:
            for line in inf:
                if line.startswith("Total Sequences"):
                    elements = line.rstrip().split("\t")
                    total_sequences = elements[1]
                    return total_sequences
        return total_sequences

    def parse_quality(self, infile, outfile):
        seq_map = {}
        read_length = 0
        with open(infile) as inf:
            check = False
            for line in inf:
                if check and line.startswith(">>END_MODULE"):
                    check = False
                if check and not line.startswith(">>END_MODULE"):
                    elements = line.rstrip().split("\t")
                    base = elements[0]
                    if base != "#Base":
                        mean = float(elements[1])
                        median = float(elements[2])
                        lower_quartile = float(
                            elements[3]
                        )  # min 20, so suggested trim length is not too strict
                        upper_quartile = float(elements[4])
                        percentile_10 = float(elements[5])
                        percentile_90 = float(elements[6])
                        if read_length < int(base):
                            read_length = int(base)
                        seq_map[int(base)] = [
                            mean,
                            median,
                            lower_quartile,
                            upper_quartile,
                            percentile_10,
                            percentile_90,
                        ]
                if line.startswith(">>Per base sequence quality"):
                    check = True

        counter = 0
        stop = False
        badass_bases = []
        for key in sorted(seq_map.keys(), reverse=True):
            if seq_map[key][2] > 20:
                stop = True
            if not stop:
                counter += 1
            else:
                if seq_map[key][2] < 20:
                    badass_bases.append(key)

        remaining = read_length - counter
        actual = 0
        if (read_length - counter) < (0.75 * read_length):
            actual = int(round(0.25 * read_length))
        else:
            actual = counter
        total_sequences = self.parse_total_sequences(infile)
        # TODO: introduce something better to write a CSV file, eg: pandas or some CSV library
        with open(outfile, "a") as outf:
            outf.write(
                "{},{},{},{},{},{},{}\n".format(
                    infile,
                    read_length,
                    counter,
                    remaining,
                    actual,
                    len(badass_bases),
                    total_sequences,
                )
            )
        return seq_map


def add_qc_parser_args(parser):
    parser.add_argument(
        "-i",
        "--input",
        dest="input",
        nargs="+",
        help="Specify input file(s) (fastqc_data.txt in fastqc folder(s))",
    )
    parser.add_argument(
        "-o", "--qc-table", dest="qc_table", help="Specify output QC table"
    )
    parser.set_defaults(func=qc_parser_command)


def qc_parser_command(args):
    open(args.qc_table, "w").write(
        "filename,read_length,suggested_trim_length,remaining_read_length,actual_trim_length,badass_bases,total_sequences\n"
    )
    p = Parser()

    for file in args.input:
        p.parse_quality(file, args.qc_table)



#!/usr/bin/env python3


import os
import subprocess


def get_input_read_count_from_star(infile):
    """Parses a star output log file to get input read counts from the fastq origin"""
    if not os.path.exists(infile):
        return -1
    with open(infile, "r") as inf:
        for line in inf:
            if line.split("|")[0].strip() == "Number of input reads":
                return int(line.split("|")[1].strip())
    return -1


def get_input_read_count_from_fastq(infile):
    """Parses input FASTQ to get read count"""
    if not os.path.exists(infile):
        return -1
    ps = subprocess.Popen(("zcat", infile), stdout=subprocess.PIPE)
    result = subprocess.check_output(("wc", "-l"), stdin=ps.stdout)
    return int(result) / 4


def get_input_read_count(infile, file_format):
    """Parses input file to get read count"""

    if file_format == "star":
        return get_input_read_count_from_star(infile)
    elif file_format == "fastq":
        return get_input_read_count_from_fastq(infile)



#!/usr/bin/env python3

"""
Junction/Spanning read pair counting

@author: TRON (PASO), BNT (URLA)
@version: 20221005
"""
import os

from logzero import logger
import pysam

from easy_fuse.misc.count_input_reads import get_input_read_count_from_star


class Requantification(object):
    """Select alignments belonging to putative fusions from an s/bam file"""

    def __init__(self, bam, output, bp_distance_threshold, input_reads_stats,):
        """Parameter initialization"""
        self.in_bam = pysam.AlignmentFile(bam, "rb")
        self.output = output
        self.bp_distance_threshold = int(bp_distance_threshold)
        self.fusion_seq_dict = {}

        # loads read counts
        assert os.path.exists(input_reads_stats) and os.path.isfile(input_reads_stats), \
            "Read stats file not found: {}".format(input_reads_stats)
        self.input_read_count = get_input_read_count_from_star(input_reads_stats)

    def quantify_read_groups(self, read_buffer, reference_base):
        """Count junction/spanning pairs on ft and wt1/2"""
        anchor_ft = 0
        anchor_wt1 = 0
        anchor_wt2 = 0
        # the read_buffer contains all reads mapping to the fusion transcript and the wildtyp backgrounds
        # the read_group is list of reads (incl multimappers) belonging to a single read pair
        # reads from this read group can map independently to ft, wt1 and wt2
        for read_group in read_buffer:
            junction_ft = 0  # 0 = no junction overlapping read, 1+ = at least one of the paired reads is a junction read
            spanning_ft = [
                0,
                0,
            ]  # left: paired read cnt fully left of the fusion bp, right: paired read cnt fulls right of the fusion bp
            junction_wt1 = 0  # like is_junction_ft but for wt1
            spanning_wt1 = [0, 0]  # like is_spanning_ft but for wt1
            junction_wt2 = 0  # like is_junction_ft but for wt2
            spanning_wt2 = [0, 0]  # like is_spanning_ft but for wt2

            # each processed read belongs to one of the three defined read_groups
            for read in read_buffer[read_group]:
                if read.reference_name.endswith("ft"):
                    junction_ft, spanning_ft, anchor_ft = self.count_junc_span(
                        read,
                        self.fusion_seq_dict[reference_base][0],
                        junction_ft,
                        spanning_ft,
                        anchor_ft,
                    )
                elif read.reference_name.endswith("wt1"):
                    junction_wt1, spanning_wt1, anchor_wt1 = self.count_junc_span(
                        read,
                        self.fusion_seq_dict[reference_base][6],
                        junction_wt1,
                        spanning_wt1,
                        anchor_wt1,
                    )
                elif read.reference_name.endswith("wt2"):
                    junction_wt2, spanning_wt2, anchor_wt2 = self.count_junc_span(
                        read,
                        self.fusion_seq_dict[reference_base][12],
                        junction_wt2,
                        spanning_wt2,
                        anchor_wt2,
                    )

            # update junction and spanning counts
            self.update_counts(reference_base, junction_ft, spanning_ft, 1)
            self.update_counts(reference_base, junction_wt1, spanning_wt1, 7)
            self.update_counts(reference_base, junction_wt2, spanning_wt2, 13)
        # set max anchor for this reference base
        self.fusion_seq_dict[reference_base][5] = anchor_ft
        self.fusion_seq_dict[reference_base][11] = anchor_wt1
        self.fusion_seq_dict[reference_base][17] = anchor_wt2

    def count_junc_span(
        self, read, breakpoint_pos, junction_count, spanning_counts, anchor
    ):
        """Identify whether a read belongs to a junction or spanning pair"""
        # read overlaps the fusion breakpoint
        if (
            breakpoint_pos >= self.bp_distance_threshold
            and read.get_overlap(
                breakpoint_pos - self.bp_distance_threshold,
                breakpoint_pos + self.bp_distance_threshold,
            )
            == 2 * self.bp_distance_threshold
        ):
            junction_count += 1
            anchor = max(
                anchor,
                min(
                    read.reference_end - breakpoint_pos,
                    breakpoint_pos - read.reference_start,
                ),
            )
        # read maps left of the fusion breakpoint
        if read.reference_end < (breakpoint_pos + self.bp_distance_threshold):
            spanning_counts[0] += 1
        # read maps right of the fusion breakpoint
        if read.reference_start > (breakpoint_pos - self.bp_distance_threshold):
            spanning_counts[1] += 1
        return junction_count, spanning_counts, anchor

    # counter_start should be 1 for ft, 6 for wt1 and 11 for wt2 (see comment in header_to_dict for details)
    def update_counts(self, reference_base, junctions, spannings, counter_start):
        """Updates the read counter in the fusion_seq_dict"""
        if spannings[0] > 0:
            self.fusion_seq_dict[reference_base][counter_start] += 1
        if spannings[1] > 0:
            self.fusion_seq_dict[reference_base][counter_start + 1] += 1
        if junctions > 0:
            self.fusion_seq_dict[reference_base][counter_start + 2] += 1
        elif spannings[0] > 0 and spannings[0] == spannings[1]:
            self.fusion_seq_dict[reference_base][counter_start + 3] += 1

    def normalize_counts_cpm(self, count):
        """Returns the CPM of a read count based on the number of original input reads"""
        if self.input_read_count == 0:
            return count
        return count / float(self.input_read_count) * 1000000.0

    def header_to_dict(self):
        """Convert bam header into a dict with keys = SN fields and values = list of breakpoint + junction/spanning counter"""
        for header_seq_id in self.in_bam.header.to_dict()["SQ"]:
            # fusion dict counter
            # [0]: breakpoint position on ft
            # [1]: gene a counts on ft
            # [2]: gene b counts on ft
            # [3]: junction read count on ft
            # [4]: spanning pairs count on ft
            # [5]: longest anchor on ft

            # [6]: breakpoint position on wt1
            # [7]: gene a counts on wt1
            # [8]: gene b counts on wt2
            # [9]: junction read count on wt1
            # [10]: spanning pairs count on wt1
            # [11]: longest anchor on wt1

            # [12]: breakpoint position on wt2
            # [13]: gene a counts on wt2
            # [14]: gene b counts on wt2
            # [15]: junction read count on wt2
            # [16]: spanning pairs count on wt2
            # [17]: longest anchor on wt2

            # currently, the header is in the format: "FTID_contextSeqHash_breakpoint_{ft,wt1,wt2}"
            # As breakpoints can be different between ft and wt, basename is everything before
            id_split = header_seq_id["SN"].split("_")
            base_name = "_".join(id_split[:-2])
            # initialize the counter list for the respective fusion context
            if base_name not in self.fusion_seq_dict:
                self.fusion_seq_dict[base_name] = [
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                ]
            if id_split[-1] == "ft":
                self.fusion_seq_dict[base_name][0] = int(id_split[-2])
            elif id_split[-1] == "wt1":
                self.fusion_seq_dict[base_name][6] = int(id_split[-2])
            elif id_split[-1] == "wt2":
                self.fusion_seq_dict[base_name][12] = int(id_split[-2])

        logger.info(
            "Found {} potential fusion sequences".format(len(self.fusion_seq_dict))
        )

    def run(self):
        """Walk linewise through a s/bam file and send reads mapping to the same fusion/wt context to counting"""
        logger.info("Starting fusion read filtering")
        count_lines = 0
        count_processed_refs = 0
        # the read buffer is a dict of lists. The dict has query names as keys and a list of corresponding reads as values.
        read_buffer = {}
        last_reference = ""
        self.header_to_dict()

        for read in self.in_bam.fetch():
            count_lines += 1
            if not last_reference in read.reference_name and count_lines > 1:
                self.quantify_read_groups(read_buffer, last_reference)
                count_processed_refs += 1
                if count_processed_refs % (len(self.fusion_seq_dict) / 10) == 0:
                    logger.info(
                        "At least {}% completed...".format(
                            count_processed_refs / (len(self.fusion_seq_dict) / 100)
                        )
                    )
                read_buffer.clear()
            if not read.query_name in read_buffer:
                read_buffer[read.query_name] = []
            read_buffer[read.query_name].append(read)
            # reference name conventions may change in the near future, but the tail must be
            # "*_breakpointPosition_ft" for the fusion transcript and "*_wt{1,2}" for the wt background
            #            if read.reference_name.endswith("ft"):
            last_reference = "_".join(read.reference_name.split("_")[:-2])
        #            else:
        #                last_reference = "_".join(read.reference_name.split("_")[:-1])

        # process the last read_buffer
        self.quantify_read_groups(read_buffer, last_reference)
        logger.info(
            "All done! Reads did not map to ~{}% of the reference input".format(
                100.0
                - (float(count_processed_refs) / (len(self.fusion_seq_dict) / 100.0))
            )
        )
        # close reading/writing stream
        self.in_bam.close()
        # write data to file. Change out_file_sep to ";" to create a csv file

        out_file_sep = ";"
        header_string = out_file_sep.join(
            [
                "ftid_plus",
                "ft_bp",
                "ft_a",
                "ft_b",
                "ft_junc",
                "ft_span",
                "ft_anch",
                "wt1_bp",
                "wt1_a",
                "wt1_b",
                "wt1_junc",
                "wt1_span",
                "wt1_anch",
                "wt2_bp",
                "wt2_a",
                "wt2_b",
                "wt2_junc",
                "wt2_span",
                "wt2_anch",
            ]
        )
        # write counts and normalized counts
        # for normalization, breakpoint and anchor must not be converted: list positions 0, 5, 10, 15
        no_norm = {0, 5, 6, 11, 12, 17}
        with open("{}.counts".format(self.output), "w") as out_counts, open(
            self.output, "w"
        ) as out_norm:
            out_counts.write("{}\n".format(header_string))
            out_norm.write("{}\n".format(header_string))
            for key in self.fusion_seq_dict:
                out_counts.write("{}{}".format(key, out_file_sep))
                out_norm.write("{}{}".format(key, out_file_sep))
                # write counts directly
                out_counts.write(
                    "{}\n".format(
                        out_file_sep.join(map(str, self.fusion_seq_dict[key]))
                    )
                )
                # normalize and write normalized counts
                self.fusion_seq_dict[key] = [
                    read_count
                    if i in no_norm
                    else self.normalize_counts_cpm(read_count)
                    for i, read_count in enumerate(self.fusion_seq_dict[key])
                ]
                # self.fusion_seq_dict[key] = map(self.normalize_counts_cpm, self.fusion_seq_dict[key])
                out_norm.write(
                    "{}\n".format(
                        out_file_sep.join(map(str, self.fusion_seq_dict[key]))
                    )
                )


def add_requantify_args(parser):
    """Add arguments for requantification to a parser"""
    parser.add_argument(
        "-i", "--input", dest="input", help="Specify input BAM file", required=True
    )
    parser.add_argument(
        "-o", "--output", dest="output", help="Specify output file", required=True
    )
    parser.add_argument(
        "-d",
        "--bp_distance",
        dest="bp_distance",
        help="Threshold of bases around the breakpoint for junction/spanning counting",
        default="3",
    )
    parser.add_argument(
        "--input-reads-stats",
        dest="input_reads_stats",
        required=True,
        help="Path to input file with reads stats",
    )
    parser.set_defaults(func=requantify_command)


def requantify_command(args):
    """Run requantification from command line"""
    requant = Requantification(args.input, args.output, args.bp_distance, args.input_reads_stats)
    requant.run()



#!/usr/bin/env python3

"""
Get predicted gene fusions from tools
Fusions are parsed into "id / fusion genes / breakpoints / supporting reads / prediction tool"
and output based on their set recurrency value. The class requires the path' of data and
output folders and writes the "Detected_Fusions.csv"

@author: Tron (PASO), BNT (URLA)
@version: 20181126
"""
import os
from typing import Dict

import logzero
from logzero import logger

from easy_fuse.fusiontoolparser_helper import *


# pylint: disable=line-too-long
#         yes they are partially, but I do not consider this to be relevant here
class FusionParser(object):
    """Get and parse results from previously run programs (fusion prediction, hla typing, expression estimation)"""

    # Initialization of parameters
    def __init__(
            self, input_fusioncatcher, input_fusioncatcher2, input_infusion, input_mapsplice,
            input_starfusion, input_soapfuse,
            output_path, sample_id, logfile):
        """Parameter initiation and work folder creation."""
        self.input_fusioncatcher = input_fusioncatcher
        self.input_fusioncatcher2 = input_fusioncatcher2
        self.input_infusion = input_infusion
        self.input_mapsplice = input_mapsplice
        self.input_starfusion = input_starfusion
        self.input_soapfuse = input_soapfuse
        self.output_path = output_path
        self.sample_id = sample_id
        logzero.logfile(logfile)

    def run(self):
        """Parse output of fusion tools and save them into output file"""

        count = 0

        logger.debug("Generating Detected Fusions table")
        detected_fusions_file = os.path.join(self.output_path, "Detected_Fusions.csv")

        with open(detected_fusions_file, "w") as all_fusions_file:
            # write header
            all_fusions_file.write(
                "BPID;Fusion_Gene;Breakpoint1;Breakpoint2;Junction_Reads;Spanning_Reads;Sample;Tool\n"
            )
            # parses all available fusion tools
            if self.input_fusioncatcher and os.path.exists(self.input_fusioncatcher):
                fusions = parse_fusioncatcher_results(self.input_fusioncatcher, self.input_fusioncatcher2)
                count += self._write_fusions(
                    all_fusions_file=all_fusions_file, fusions_dict=fusions, tool="fusioncatcher")
            if self.input_infusion and os.path.exists(self.input_infusion):
                fusions = parse_infusion_results(self.input_infusion)
                count += self._write_fusions(
                    all_fusions_file=all_fusions_file, fusions_dict=fusions, tool="infusion")
            if self.input_mapsplice and os.path.exists(self.input_mapsplice):
                fusions = parse_mapsplice_results(self.input_mapsplice)
                count += self._write_fusions(
                    all_fusions_file=all_fusions_file, fusions_dict=fusions, tool="mapsplice")
            if self.input_starfusion and os.path.exists(self.input_starfusion):
                fusions = parse_starfusion_results(self.input_starfusion)
                count += self._write_fusions(
                    all_fusions_file=all_fusions_file, fusions_dict=fusions, tool="starfusion")
            if self.input_soapfuse and os.path.exists(self.input_soapfuse):
                fusions = parse_soapfuse_results(self.input_soapfuse)
                count += self._write_fusions(
                    all_fusions_file=all_fusions_file, fusions_dict=fusions, tool="soapfuse")

        logger.info(
            "Wrote {0} detected fusion genes to {1}.".format(
                count, detected_fusions_file
            )
        )

    def _write_fusions(self, fusions_dict: Dict, all_fusions_file, tool):
        count = 0

        # write fusions for this tool to a file
        tool_res_file = os.path.join(self.output_path, tool.lower() + ".csv")
        with open(tool_res_file, "w") as tool_outf:
            tool_outf.write(
                "bpid;fusion_gene;breakpoint1;breakpoint2;junc_reads;span_reads\n"
            )
            for key in fusions_dict:
                tool_outf.write("{};{}\n".format(key, ";".join(fusions_dict[key])))

        # add fusions from this tool to all fusions file
        for bpid in fusions_dict:
            count += 1
            all_fusions_file.write(
                "{};{};{};{}\n".format(
                    bpid, ";".join(fusions_dict[bpid]), self.sample_id, tool
                )
            )
        return count


def add_fusion_parse_args(parser):

    parser.add_argument(
        "--input-fusioncatcher",
        dest="input_fusioncatcher",
        help="Input file with results from fusion catcher (ie: summary_candidate_fusions.txt).",
    )
    parser.add_argument(
        "--input-fusioncatcher2",
        dest="input_fusioncatcher2",
        help="Input file with results from fusion catcher (ie: final-list_candidate-fusion-genes.txt).",
    )
    parser.add_argument(
        "--input-starfusion",
        dest="input_starfusion",
        help="Input file with results from star fusion (ie: star-fusion.fusion_predictions.tsv).",
    )
    parser.add_argument(
        "--input-infusion",
        dest="input_infusion",
        help="Input file with results from infusion (ie: final_fusion_genes).",
    )
    parser.add_argument(
        "--input-mapsplice",
        dest="input_mapsplice",
        help="Input file with results from mapsplice (ie: fusions_well_annotated.txt).",
    )
    parser.add_argument(
        "--input-soapfuse",
        dest="input_soapfuse",
        help="Input file with results from soapfuse "
             "(ie: final_fusion_genes/${SAMPLE}/${SAMPLE}.final.Fusion.specific.for.genes).",
    )
    parser.add_argument(
        "-o",
        "--output",
        dest="output_path",
        help="Specify the fusion output folder of the sample.",
        required=True,
    )
    parser.add_argument(
        "-s",
        "--sample",
        dest="sample",
        help="Specify the sample to process.",
        required=True,
    )
    parser.add_argument(
        "-l", "--logger", dest="logger", help="Logging of processing steps", default=""
    )
    parser.set_defaults(func=fusion_parser_command)


def fusion_parser_command(args):

    # checks that at least one fusion prediction tool is provided
    assert args.input_fusioncatcher or args.input_starfusion or args.input_infusion or args.input_mapsplice or \
           args.input_soapfuse, "No fusion prediction tool results provided."

    fusion_parser = FusionParser(
        input_fusioncatcher=args.input_fusioncatcher,
        input_fusioncatcher2=args.input_fusioncatcher2,
        input_starfusion=args.input_starfusion,
        input_infusion=args.input_infusion,
        input_mapsplice=args.input_mapsplice,
        input_soapfuse=args.input_soapfuse,
        output_path=args.output_path,
        sample_id=args.sample,
        logfile=args.logger
    )
    fusion_parser.run()



#!/usr/bin/env python

"""
First truncates the header of a multi fasta file to the first whitespace
Second splits the modified file into separate fasta files
The separated chromosomes are at least mandatory for mapsplice2 and fetchdata
Truncating of the headers is also mandatory for mapsplice2 and in general
a good idea to avoid any kind of parsing issues

@author: BNT (URLA)
@version: 20180131
"""

from __future__ import print_function

import os.path
from argparse import ArgumentParser

from logzero import logger


class GenomeRefParser(object):
    """Initialize class variables"""

    def __init__(self, in_fasta, out_dir):
        self.in_fasta = in_fasta
        self.out_dir = out_dir

    def short_header(self):
        """Shorten the header in a multi fasta file to the first whitespace"""
        out_fasta = "{}_SH.fa".format(self.in_fasta.rsplit(".", 1)[0])
        with open(self.in_fasta, "r") as fasin, open(out_fasta, "w") as fasout:
            for line in fasin:
                if line.startswith(">"):
                    logger.info(
                        'Changing header from "{0}" to "{1}"'.format(
                            line.rstrip("\n"), line.split()[0]
                        )
                    )
                    fasout.write("{}\n".format(line.split()[0]))
                else:
                    fasout.write(line)
        self.in_fasta = out_fasta

    def split(self):
        """Performing splitting of multi fasta file"""
        fasout = None
        with open(self.in_fasta, "r") as fasin:
            for line in fasin:
                if line.startswith(">"):
                    if fasout:
                        fasout.close()
                    out_file_name = os.path.join(
                        self.out_dir, "{}.fa".format(line.split()[0][1:])
                    )
                    logger.info(
                        'Writing fasta record "{0}" to file "{1}"'.format(
                            line.rstrip("\n"), out_file_name
                        )
                    )
                    fasout = open(out_file_name, "w")
                fasout.write(line)
            if fasout:
                fasout.close()


def main():
    """Parse command line arguments and start script"""
    parser = ArgumentParser(description="Generate mapping stats for fusion detection")
    parser.add_argument(
        "-i", "--input", dest="input", help="multi fasta file", required=True
    )
    parser.add_argument(
        "-o", "--output", dest="output", help="Output directory", required=True
    )
    args = parser.parse_args()

    grp = GenomeRefParser(args.input, args.output)
    grp.short_header()
    grp.split()


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

"""
Input/Output method collection
- Create folder and grant permission
- Read fastq files from directories
- get basename from full path

@author: Tron (PASO), BNT (URLA)
@version: 20181126
"""

import os
import re
import stat

from logzero import logger


def create_folder(path):
    """This function creates a specified folder, if not already existing and grants the respective permission."""
    if not os.path.exists(path):
        os.makedirs(path)
        grant_permissions(
            path, stat.S_IRWXU | stat.S_IRWXG | stat.S_IROTH | stat.S_IXOTH
        )
        logger.debug("Created folder {}".format(path))
    else:
        logger.debug("Folder {} already exists".format(path))


def grant_permissions(path, mode):
    """This function grants specific permissions for a given folder or file."""
    for _, dirs, files in os.walk(path, topdown=False):
        for dir in dirs:
            os.chmod(dir, mode)
        for file in files:
            os.chmod(file, mode)


def get_fastq_files(input_paths):
    """Load fastq files, check file name consistency and return tuple of left/right pairs"""
    # load fastq files from the provide path to a list
    fastqs = []
    for path in input_paths:
        if os.path.isdir(path):
            files = os.listdir(path)
            for filename in files:
                file_path = os.path.join(path, filename)
                if os.path.isfile(file_path) and filename.endswith(
                    (".fastq.gz", ".fq.gz", ".fq", ".fastq")
                ):
                    if not filename.endswith(".gz"):
                        logger.info(
                            "Warning: Fastq files are supposed to be gzipped! Skipping record {}!".format(
                                filename
                            )
                        )
                        continue
                    fastqs.append(file_path)
                elif os.path.isdir(file_path):
                    fastqs_tmp = get_fastq_files([file_path])
                    fastqs.extend(fastqs_tmp)
        elif os.path.isfile(path) and path.endswith(
            (".fastq.gz", ".fq.gz", ".fq", ".fastq")
        ):
            fastqs.append(path)
    return fastqs


def pair_fastq_files(fastqs):
    left = []
    right = []
    # iterate over the sorted list of file names and check for left/right pair file
    logger.info("Going to process the following read files...")
    for i, fq_file in enumerate(sorted(fastqs)):
        logger.info("Fastq file {0}: {1}".format(i, fq_file))
        try:
            # Search for 1 or 2 between "_R|_" and "_|.f" in the basename of the file
            filename = os.path.basename(fq_file)
            matches = re.search("(.*)(_R?[1-2])(_|[.])", filename)
            if matches:
                sample_id = matches.group(1)
                forrev = matches.group(2)
            else:
                raise ValueError(
                    "File failed to match name pattern: {}".format(filename)
                )

            if forrev == "_1" or forrev == "_R1":
                left.append((sample_id, fq_file))
            elif forrev == "_2" or forrev == "_R2":
                right.append((sample_id, fq_file))

        except AttributeError:
            forrev = "-1"

    # Check whether file names match between paired files
    sample_list = []
    # if right:
    for sample_id_a, fq_file_a in left:
        found = False
        for sample_id_b, fq_file_b in right:
            if sample_id_a == sample_id_b:
                found = True
                sample_list.append((sample_id_a, fq_file_a, fq_file_b))
                break

        if not found:
            logger.info("Could not find a matching FASTQ for {}!".format(sample_id_a))
            sample_list.append((sample_id_a, fq_file_a, ""))

    return sample_list



#!/usr/bin/env python3

"""
Read input bam file generated by star
Star must have been run with "--outSAMmultNmax 1" and "--chimSegmentMin X"
Aligned pairs are filtered according to their likely of being chimeric with 4 subsequent filter:
    1) the pair is defined as chimeric by star
    2) at least one of the reads in the pair is unmapped
    3) the pair is not a proper pair (unexpected distance)
    4) 10 or more bases of at least one of the reads in the pair are soft-clipped
Multi-mappers are currently discarded as long as they are not belonging to filter "1"
All pairs that are not either "1-4", are defined as unlikely chimeric
A log file is generated after successful completion which contains information about:
    1) number of input vs filtered reads
    2) runtime/speed of the processing
    3) filter counts
    4) 2 qc tests for fast checking that counted pairs are as expected

@author: Tron (PASO), BNT (URLA)
@version: 20181126
"""

import sys
import time

import logzero
import pysam
from logzero import logger


class FusionReadFilter(object):
    """Select alignments belonging to putative fusions from an s/bam file"""

    def __init__(self, bam, output):
        """Parameter initialization"""
        self.bam_file = bam
        self.in_bam = pysam.AlignmentFile(bam, "rb")
        self.filtered_bam = pysam.AlignmentFile(output, "wb", template=self.in_bam)
        # a list of ints is used to count pairs assigned to different filters
        # 0: count_input_alignments
        # 1: count_input_pairs
        # 2: count_filtered_pairs
        # 3: count_multimapped
        # 4: count_star_chimeric_alignments
        # 5: count_qcd_alignments
        # 6: count_unmapped
        # 7: count_10bp_s_clip
        # 8: count_proper_pair
        self.counter = [0, 0, 0, 0, 0, 0, 0, 0, 0]
        self.last_time = 0
        logzero.logfile("{}.fusionReadFilterLog".format(output))

    def classify_pair(self, read1, read2, read1_flag, read2_flag, member_count):
        """Classify a read pairs into different groups"""
        self.counter[1] += 1

        if read1 is None or read2 is None:
            return False

        # count star chimeric
        if member_count > 2:
            self.counter[4] += 1
            return True
        # count multimapping reads
        if read1.get_tag("NH") > 1 or read2.get_tag("NH") > 1:
            self.counter[3] += 1
            return False
        # count pairs containing unmapped mates
        if read1_flag & 0x4 or read2_flag & 0x4:
            self.counter[6] += 1
            return True
        # check if the pair is a "proper pair" (both mapped, adequat distance)
        if not read1_flag & 0x2 or not read2_flag & 0x2:
            self.counter[8] += 1
            return True
        # check if 10 (default) or more soft-clippings (S) are in the alignment
        if (
            read1.query_alignment_length + 10 <= read1.query_length
            or read2.query_alignment_length + 10 <= read2.query_length
        ):
            self.counter[7] += 1
            return True
        return False

    def run(self):
        """Walk linewise through a s/bam file and send proper read pairs to classification"""
        logger.info("Starting fusion read filtering")
        read1 = read2 = None
        read1_flag = read2_flag = None
        self.last_time = time.time()
        last_query = ""
        count_current_query_member = 0
        # read filtering works as follows:
        # urla: I wrote the processing in a way, that implementing the analyses of multi-mapping reads should be straight forward
        #       For this, however, one would need to check in the group of multi-mapping alignments whether there is no possible "normal mapping"
        #       I could so far think of two possible solutions: (1) running "classify_pair" on all possible combinations of the multimapping pairs,
        #       until a "normal mapping" was found. If this is never the case, the pair is written to the filtered out. This, however, will have
        #       a strong effect on runtimes! (2) with the chimeric multi-mapping setting in the most recent star versions. This should actually
        #       work directly as star chimeric classification outranks multi-mapping disposal. It has, however, not been throughly evaluated by the star community

        for read in self.in_bam.fetch(until_eof=True):
            # iterate through alignments as they appear in the file (this is mandatory because
            # (a) we cannot create an index,
            # (b) want to include unmapped reads and
            # (c) have many references in the header during 2nd filtering)
            self.counter[0] += 1
            read_flag = read.flag

            if last_query != read.query_name and self.counter[0] > 1:
                if self.classify_pair(
                    read1, read2, read1_flag, read2_flag, count_current_query_member
                ):
                    self.filtered_bam.write(read1)
                    self.filtered_bam.write(read2)
                    self.counter[2] += 1
                read1 = read2 = None
                read1_flag = read2_flag = None
                count_current_query_member = 0

            count_current_query_member += 1
            # ignore all alignments which are either supplemental, vendor qc'd, secondary or duplicates
            # urla: duplicate flagging/removal prior to running this script would most probably lead to several errors
            #       Nevertheless, it is a very fast check and I would also not recommend that someone does deduplication on rna-seq data!
            if read_flag > 255:
                self.counter[5] += 1
            else:
                # urla: the following commented version should be fine, but the other one is still kept in order to be more error-aware
                #                if read.is_read1:
                #                    read1 = read
                #                else:
                #                    read2 = read
                if not read1 and read.is_read1:
                    read1 = read
                    read1_flag = read_flag
                elif not read2 and read.is_read2:
                    read2 = read
                    read2_flag = read_flag
                else:
                    logger.error(
                        "Neither r1 nor r2??? Read: {0}; R1: {1}; R2: {2}; bamLine: {3}".format(
                            read, read1, read2, self.counter[0]
                        )
                    )
                    sys.exit(1)
            last_query = read.query_name

        # once EOF is reached, the very last pair has to be classified additionally
        if self.classify_pair(
            read1, read2, read1_flag, read2_flag, count_current_query_member
        ):
            self.filtered_bam.write(read1)
            self.filtered_bam.write(read2)
            self.counter[2] += 1

        # close reading/writing stream
        self.in_bam.close()
        self.filtered_bam.close()
        self.print_stats()
        logger.info("Finished fusion read filtering")

    def print_stats(self):
        """print collected statistics to the log file"""
        this_time = time.time()
        time_taken = this_time - self.last_time
        time_taken_1m = 0
        if self.counter[0]:
            time_taken_1m = float(time_taken * 1000000) / self.counter[0]
        self.last_time = this_time

        time_taken_2m = 0
        if self.counter[1]:
            time_taken_2m = float(self.counter[2]) / self.counter[1]

        logger.info(
            "Processed {0} alignments, {1} of {2} pairs remained after filtering ({3:.2%}) ({4:.2f}s / 1M alignments; {5:.2f}s in total)".format(
                self.counter[0],
                self.counter[2],
                self.counter[1],
                time_taken_2m,
                time_taken_1m,
                time_taken,
            )
        )

        qc2 = False
        if (self.counter[4] == self.counter[5]) and (
            (self.counter[0] - self.counter[5]) * 0.5 == self.counter[1]
        ):
            qc2 = True

        # 0: count_input_alignments
        # 1: count_input_pairs
        # 2: count_filtered_pairs
        # 3: count_multimapped
        # 4: count_star_chimeric_alignments
        # 5: count_qcd_alignments
        # 6: count_unmapped
        # 7: count_10bp_s_clip
        # 8: count_proper_pair
        logger.info(
            "Star_chimeric (chim alignment from star):\t{} pairs (filtered)".format(
                self.counter[4]
            )
        )
        logger.info(
            "QC'd (additional Star_chimeric alignment):\t{} alignments (included in above)".format(
                self.counter[5]
            )
        )
        logger.info(
            "Multimapped (1 < x <= 100 equal mappings):\t{} pairs (discarded)".format(
                self.counter[3]
            )
        )
        logger.info(
            "Unmapped (no mapping or >100 multi map):\t{} pairs (filtered)".format(
                self.counter[6]
            )
        )
        logger.info(
            "No proper pair (unexpected read distance):\t{} pairs (filtered)".format(
                self.counter[8]
            )
        )
        logger.info(
            "10bp_s_clip (>9bp soft-clipped in cigar):\t{} pairs (filtered)".format(
                self.counter[7]
            )
        )
        logger.info(
            'Unlikely chimeric ("normal" mappings): \t{} pairs (discarded)'.format(
                self.counter[1]
                - self.counter[4]
                - self.counter[3]
                - self.counter[6]
                - self.counter[8]
                - self.counter[7]
            )
        )
        logger.info("Filter QC2 (QC'd alignments are chimeric):\t{}".format(qc2))


def add_read_filter_args(parser):
    parser.add_argument("-i", "--input", dest="input", help="Specify input BAM file. Must be sorted by query name")
    parser.add_argument("-o", "--output", dest="output", help="Specify output prefix")
    parser.set_defaults(func=read_filter_command)


def read_filter_command(args):
    read_filter = FusionReadFilter(args.input, args.output)
    read_filter.run()



#!/usr/bin/env python


import math
import subprocess


def get_sequence_count_and_len(fasta_file):
    """Returns the number of sequences and the total bases from a fasta file"""
    seq_num = 0
    total_bases = 0
    with open(fasta_file) as fasta:
        for line in fasta:
            if line[0] == ">":
                seq_num += 1
            else:
                total_bases += len(line.rstrip())
    return seq_num, total_bases


def get_pseudo_genome_adjustments_for_star(fasta_file):
    """Return the genome size of an associated fasta file"""
    (seq_num, genome_size) = get_sequence_count_and_len(fasta_file)
    star_genome_chr_bin_n_bits = min(18, int(math.log(genome_size / seq_num, 2)))
    star_genome_sa_index_n_bases = min(14, int(math.log(genome_size, 2) / 2 - 1)) - 2

    return str(star_genome_chr_bin_n_bits), str(star_genome_sa_index_n_bases)


def run(input_fasta, genome_dir, threads, star_binary):
    (chr_bin_n_bits, sa_index_n_bases) = get_pseudo_genome_adjustments_for_star(
        input_fasta
    )

    cmd_starindex = "{0} --runMode genomeGenerate --runThreadN {1} --limitGenomeGenerateRAM 48000000000 --genomeChrBinNbits {2} --genomeSAindexNbases {3} --genomeDir {4} --genomeFastaFiles {5}".format(
        star_binary, threads, chr_bin_n_bits, sa_index_n_bases, genome_dir, input_fasta
    )

    subprocess.run(cmd_starindex.split(" "))


def add_star_custom_index_args(parser):
    parser.add_argument(
        "-i",
        "--input_fasta",
        dest="input_fasta",
        help="Specify input fasta file",
        required=True,
    )
    parser.add_argument(
        "-o",
        "--genome_dir",
        dest="genome_dir",
        help="Specify path to output genome dir",
        required=True,
    )
    parser.add_argument(
        "-t",
        "--threads",
        dest="threads",
        help="Specify amount of threads to be used during runtime",
        default=1,
    )
    parser.add_argument(
        "-b",
        "--star_binary",
        dest="star_binary",
        help="Specify path to STAR binary for execution",
        default="STAR",
    )
    parser.set_defaults(func=star_custom_index_command)


def star_custom_index_command(args):
    run(args.input_fasta, args.genome_dir, args.threads, args.star_binary)



#!/usr/bin/env python3

"""
@author: BNT (URLA), TRON (PASO)
@version: 20190118
"""
import os
import os.path
import time

import easy_fuse
import pkg_resources
from logzero import logger
import pandas as pd
from easy_fuse.misc import queueing
from easy_fuse.misc.count_input_reads import get_input_read_count_from_star


class FusionSummary(object):
    """Collect stats of the run and write them to file"""

    def __init__(
            self,
            input_fusions,
            input_fusion_context_seqs,
            input_requant_cpm,
            input_requant_counts,
            input_reads_stats,
            input_model_pred,
            output_folder,
            requant_mode: str,
            model_pred_threshold: float,
            fusion_tools: list
    ):

        assert os.path.exists(input_fusions) and os.path.isfile(input_fusions), \
            "Fusion file not found: {}".format(input_fusions)
        self.input_fusions = input_fusions
        assert os.path.exists(input_fusion_context_seqs) and os.path.isfile(input_fusion_context_seqs), \
            "Fusion context seq file not found: {}".format(input_fusion_context_seqs)
        self.input_fusion_context_seqs = input_fusion_context_seqs
        assert os.path.exists(input_requant_cpm) and os.path.isfile(input_requant_cpm), \
            "Requant CPM file not found: {}".format(input_requant_cpm)
        self.input_requant_cpm = input_requant_cpm
        assert os.path.exists(input_requant_counts) and os.path.isfile(input_requant_counts), \
            "Requant counts file not found: {}".format(input_requant_counts)
        self.input_requant_counts = input_requant_counts
        assert os.path.exists(input_reads_stats) and os.path.isfile(input_reads_stats), \
            "Read stats file not found: {}".format(input_reads_stats)
        self.input_reads_stats = input_reads_stats
        assert os.path.exists(input_model_pred) and os.path.isfile(input_model_pred), \
            "Prediction model file not found: {}".format(input_model_pred)
        self.input_model_pred = input_model_pred
        assert os.path.exists(output_folder) and os.path.isdir(output_folder), \
            "Output folder not found: {}".format(output_folder)
        self.output_folder = output_folder
        self.model_pred_threshold = model_pred_threshold
        self.fusion_tools = fusion_tools

        # loads read counts
        assert os.path.exists(input_reads_stats) and os.path.isfile(input_reads_stats), \
            "Read stats file not found: {}".format(input_reads_stats)
        self.input_read_count = get_input_read_count_from_star(input_reads_stats)

        # loads more data
        self.data = self.load_data(requant_mode=requant_mode)

        pd.options.mode.chained_assignment = None
        # TODO: we are not using this blacklist, reconsider if we need this
        self.blacklist = []

    def load_data(self, requant_mode):
        """Join the three data tables context_seq, detected_fusions and requantification"""
        # define path' to the context seq, detected fusion and re-quantification files

        context_data = pd.read_csv(self.input_fusion_context_seqs, sep=";")
        # append key for later join with requant data
        context_data["ftid_plus"] = (
                context_data["FTID"] + "_" + context_data["context_sequence_id"]
        )

        # read the original append CPM values from individual fusion prediction tools to context data
        detect_data = pd.read_csv(self.input_fusions, sep=";")
        logger.info(
            "Appending normalized fusion counts from {} to the data table.".format(
                ",".join(self.fusion_tools)
            )
        )

        context_data = self.append_tool_cnts_to_context_file(
            context_data, detect_data
        )

        logger.info(
            "Appending (normalized) requantification counts from filtered and original mapping to the data table."
        )
        # perform subsequent joins on ftid_plus
        context_data.set_index("ftid_plus", inplace=True)

        # read and append normalized (cpm) and raw (counts) requantification data to context data
        # for mode in requant_mode:
        requant_cpm_data = pd.read_csv(self.input_requant_cpm, sep=";")
        context_data = context_data.join(
            requant_cpm_data.set_index("ftid_plus"), how="left"
        )
        requant_cnt_data = pd.read_csv(self.input_requant_counts, sep=";")
        context_data = context_data.join(
            requant_cnt_data.set_index("ftid_plus"),
            lsuffix="_{}".format(requant_mode),
            rsuffix="_cnt_{}".format(requant_mode),
            how="left",
        )

        return context_data.fillna(0)

    def run(self, model_predictions):
        """Execute individual methods"""

        # 0:type; 1:boundary; 2:frame; 3:pepseq; 4:counts; 5:blacklist; 6:prediction;
        # 7:exonError; 8:unfiltered; 9:allFilter; 10:allButPredFilter

        start_time = time.time()

        # check whether output already exists
        cols_to_aggregate_on = ["BPID", "context_sequence_id", "FTID"]

        # in comparison to the previous implementation, pre-existing results will always be overwritten as it turned out that the
        # previous implementation is to error prone...
        joined_table_1b = self.data.groupby(
            by=cols_to_aggregate_on, sort=False, as_index=False
        )
        ranked_fusions_file = os.path.join(self.output_folder, "fusions.csv")
        joined_table_1b.agg(self.custom_data_aggregation).to_csv(
            ranked_fusions_file, sep=";", index=False
        )

        joined_table_1 = pd.read_csv(ranked_fusions_file, sep=";")

        if model_predictions and os.path.exists(ranked_fusions_file) and os.path.isfile(ranked_fusions_file):
            model_path = os.path.abspath(self.input_model_pred)

            # append prediction scores based on pre-calculated model
            predicted_fusions_file = os.path.join(self.output_folder, "fusions.pass.csv")
            cmd_model = "{0} --fusion_summary {1} --model_file {2} --prediction_threshold {3} --output {4}".format(
                pkg_resources.resource_filename(
                    easy_fuse.__name__, "resources/R/R_model_prediction.R"
                ),
                ranked_fusions_file,
                model_path,
                self.model_pred_threshold,
                predicted_fusions_file,
            )

            queueing.submit(
                "", cmd_model.split(" "), "", "", "", "", "", "", "", "", "", "none"
            )
            # re-read the table with prediction for filter counting
            # urla - note: there is probably a more elegant way using getattr/setattr but I'm not at all familiar with its pros/cons
            
            joined_table_1 = pd.read_csv(
                predicted_fusions_file, sep=";"
            )

        joined_table_1.set_index(keys=cols_to_aggregate_on, drop=False, inplace=True)
        time_taken = time.time() - start_time
        logger.info(
            "Done. Processing time: {0:.2f} s".format(
                time_taken
            )
        )

        return self.count_records(joined_table_1, False, "F1N", model_predictions=model_predictions)


    def count_records(self, pd_data_frame, is_merged, name, model_predictions):
        """stub"""
        logger.info("Processing: {}".format(name))
        #        blacklist = ["HLA", "IG"]
        # list of sets to store (un)filtered fusion gene names
        # 0:unfiltered; 1:type; 2:boundary; 3:frame; 4:pepseq; 5:counts;
        # 6:blacklist; 7:prediction; 8:allFilter; 9:allButPredFilter
        # this list is initialized with empty sets which are either replaced with data or stay empty
        # this guarantees that filters are always at the same position in the list, even if they are empty
        fusion_gene_set_list = [set() for i in range(10)]
        # the dataframe may be empty if (a) no fusions were predicted or (b) no fusion remained after replicate merging
        if pd_data_frame.empty:
            return map(len, fusion_gene_set_list), set()
        df_rep = pd_data_frame
        # try to correct annotation issues
        ftid_dict = {}
        for i in df_rep.index:
            ftid = df_rep.loc[i, "FTID"]
            if ftid not in ftid_dict:
                ftid_dict[ftid] = [df_rep.loc[i, "Fusion_Gene"]]
            else:
                ftid_dict[ftid].append(df_rep.loc[i, "Fusion_Gene"])

        for ftid in ftid_dict:
            # check if multiple fusion_gene strings are associated with the ftid
            if len(ftid_dict[ftid]) > 1:
                # check if there is one of the multiple fusion_gene records where both partners correspond to the ftid
                for fusion_gene in ftid_dict[ftid]:
                    try:
                        left_fusion, right_fusion = fusion_gene.split("_")
                        if left_fusion in ftid and right_fusion in ftid:
                            ftid_dict[ftid] = fusion_gene
                            break
                    except ValueError:
                        logger.warning(
                            "Fusion gene is in an unexpected format (If you have gene names containing underscores, please contact the authors!)"
                        )
                # if none of the fusion_gene records fit to the ftid, take the shortest and print a warning
                if type(ftid_dict[ftid]) == list:
                    logger.warning(
                        'Unresolvable annotation bias detected between ftid "{0}" and fusion_gene records "{1}"'.format(
                            ftid, ftid_dict[ftid]
                        )
                    )
                    ftid_dict[ftid] = min(ftid_dict[ftid], key=len)
                    logger.warning(
                        "Chosing {} as fusion_gene record which might be wrong.".format(
                            ftid_dict[ftid]
                        )
                    )
            else:
                ftid_dict[ftid] = ftid_dict[ftid][0]

        # create sets of fusion gene names according to different fiters
        fusion_gene_set_list[1] = set(
            [
                ftid_dict[x]
                for x in df_rep[df_rep["type"].astype(str) != "cis_near"]["FTID"]
            ]
        )  # exclude cis near events as they are likely read through
        fusion_gene_set_list[2] = set(
            [
                ftid_dict[x]
                for x in df_rep[df_rep["exon_boundary"].astype(str) == "both"]["FTID"]
            ]
        )  # allow only fusions where the breakpoints are on exon boundaries in BOTH fusion partners
        fusion_gene_set_list[3] = set(
            [
                ftid_dict[x]
                for x in df_rep[df_rep["frame"].astype(str) != "no_frame"]["FTID"]
            ]
        )  # exclude no frame fusions
        # urla - note: I'm not 100% why str.isalpha is working everywhere I tested, but not on our dev servers... This way, it works, although I thought the pandas str method would this already...
        fusion_gene_set_list[4] = set(
            [
                ftid_dict[x]
                for x in df_rep[
                    df_rep["neo_peptide_sequence"].astype(str).str.isalpha()
                ]["FTID"]
            ]
        )  # the neo peptide sequence must be an alphabetic sequence of at least length 1
        if is_merged:
            junc_cnt_col = "ft_junc_cnt_org_1"
            if not junc_cnt_col in list(df_rep):
                junc_cnt_col = "ft_junc_cnt_best_1"
                if not junc_cnt_col in list(df_rep):
                    junc_cnt_col = "ft_junc_cnt_fltr_1"
            fusion_gene_set_list[5] = set(
                [
                    ftid_dict[x]
                    for x in df_rep[
                        df_rep[junc_cnt_col]
                        + df_rep[junc_cnt_col.replace("junc", "span")]
                        + df_rep[junc_cnt_col.replace("_1", "_2")]
                        + df_rep[
                            junc_cnt_col.replace("junc", "span").replace("_1", "_2")
                        ]
                        > 0
                    ]["FTID"]
                ]
            )  # at least 1 junction or spanning counts on the fusion transcript in either replicate
        else:
            junc_cnt_col = "ft_junc_cnt_org"
            if not junc_cnt_col in list(df_rep):
                junc_cnt_col = "ft_junc_cnt_best"
                if not junc_cnt_col in list(df_rep):
                    junc_cnt_col = "ft_junc_cnt_fltr"
            fusion_gene_set_list[5] = set(
                [
                    ftid_dict[x]
                    for x in df_rep[
                        df_rep[junc_cnt_col]
                        + df_rep[junc_cnt_col.replace("junc", "span")]
                        > 0
                    ]["FTID"]
                ]
            )  # at least 1 junction or spanning counts on the fusion transcript
        # both fusion partners are not allowed to start with anything mentioned in the "blacklisted" list (single gene/gene family exclusion)
        # and the fusion gene is not allowed to be identical to a string in the blacklist (gene pair exclusion)
        fusion_gene_set_list[6] = set(
            [
                ftid_dict[z]
                for z in df_rep[
                    [
                        all(
                            [
                                not (
                                    y.startswith(tuple(self.blacklist))
                                    or "_".join(x) in self.blacklist
                                )
                                for y in x
                            ]
                        )
                        for x in df_rep["Fusion_Gene"].str.split("_")
                    ]
                ]["FTID"]
            ]
        )
        if model_predictions:
            if is_merged:
                fusion_gene_set_list[7] = set(
                    [
                        ftid_dict[x]
                        for x in df_rep[
                            (df_rep["prediction_class_1"].astype(str) == "positive")
                            | (df_rep["prediction_class_2"].astype(str) == "positive")
                        ]["FTID"]
                    ]
                )
                # the random forrest model must have classified this at least once as "posititve" in either replicate
            else:
                fusion_gene_set_list[7] = set(
                    [
                        ftid_dict[x]
                        for x in df_rep[
                            df_rep["prediction_class"].astype(str) == "positive"
                        ]["FTID"]
                    ]
                )
                # the random forrest model must have classified this at least once as "posititve"
        fusion_gene_filter_wo_pred = set(ftid_dict.values())
        # intersect with all but predictions
        for i in range(1, 7):
            fusion_gene_filter_wo_pred = fusion_gene_filter_wo_pred.intersection(
                fusion_gene_set_list[i]
            )
        fusion_gene_filter_all = fusion_gene_filter_wo_pred.intersection(
            fusion_gene_set_list[7]
        )

        fusion_gene_set_list[0] = set(ftid_dict.values())
        fusion_gene_set_list[8] = fusion_gene_filter_all
        fusion_gene_set_list[9] = fusion_gene_filter_wo_pred

        if model_predictions:
            return map(len, fusion_gene_set_list), fusion_gene_filter_all
        return map(len, fusion_gene_set_list), fusion_gene_filter_wo_pred

    def append_tool_cnts_to_context_file(
        self, context_data, detected_fusions
    ):
        """Add data to context seq table"""
        # Init new columns
        for tool in self.fusion_tools:
            context_data["{}_detected".format(tool.lower())] = 0
            context_data["{}_junc".format(tool.lower())] = 0
            context_data["{}_span".format(tool.lower())] = 0
        context_data["tool_count"] = 0
        context_data["tool_frac"] = 0.0

        # iterate over context seq data
        for i in context_data.index:
            # get tool prediction for each BPID
            tmp_slice = detected_fusions.loc[
                detected_fusions["BPID"] == context_data["BPID"][i]
            ]
            # check for each tool, whether a prediction was made by it and update the respective columns
            for tool in self.fusion_tools:
                # pre-set values to 0 or NA indicating no prediction with this tool
                # line wise initialization shouldn't be required as it was initialized with this values in the beginning
                for j in tmp_slice.index:
                    if tool == tmp_slice.loc[j, "Tool"]:
                        context_data.loc[i, "{}_detected".format(tool.lower())] = 1
                        context_data.loc[
                            i, "{}_junc".format(tool.lower())
                        ] = self.normalize_counts_cpm(
                            tmp_slice.loc[j, "Junction_Reads"]
                        )
                        context_data.loc[
                            i, "{}_span".format(tool.lower())
                        ] = self.normalize_counts_cpm(
                            tmp_slice.loc[j, "Spanning_Reads"]
                        )
                        context_data.loc[i, "tool_count"] += 1
            context_data.loc[i, "tool_frac"] = float(
                context_data.loc[i, "tool_count"]
            ) / float(len(self.fusion_tools))
        return context_data

    def load_blacklist(self, blacklist_file):
        """Read blacklist data from file"""
        if blacklist_file is not None:
            with open(blacklist_file, "r") as bl_file:
                for line in bl_file:
                    if line.startswith("#"):
                        continue
                    self.blacklist.append(line.strip())

    def normalize_counts_cpm(self, count):
        """Returns the CPM of a read count based on the number of original input reads"""
        if self.input_read_count == 0:
            return count
        return count / float(self.input_read_count) * 1000000.0

    @staticmethod
    def custom_data_aggregation(pd_series):
        """Removes duplicates from a pandas series fields and creates a joined string from the resulting list"""
        # urla - note: this is an adaptation from https://stackoverflow.com/questions/480214/how-do-you-remove-duplicates-from-a-list-whilst-preserving-order
        #              in a way that it does not return a list but a string to print from the list
        seen = set()
        seen_add = seen.add
        return ";".join(
            str(e) for e in [x for x in pd_series if not (x in seen or seen_add(x))]
        )


def add_summarize_data_args(parser):

    parser.add_argument(
        "--input-fusions",
        dest="input_fusions",
        required=True,
        help="Path to input file with fusions",
    )
    parser.add_argument(
        "--input-fusion-context-seqs",
        dest="input_fusion_context_seqs",
        required=True,
        help="Path to input file with fusion context sequences",
    )
    parser.add_argument(
        "--input-requant-cpm",
        dest="input_requant_cpm",
        required=True,
        help="Path to input file with requant CPM",
    )
    parser.add_argument(
        "--input-requant-counts",
        dest="input_requant_counts",
        required=True,
        help="Path to input file with requant counts",
    )
    parser.add_argument(
        "--input-reads-stats",
        dest="input_reads_stats",
        required=True,
        help="Path to input file with reads stats",
    )
    parser.add_argument(
        "--input-model-pred",
        dest="input_model_pred",
        required=True,
        help="Path to input file with prediction model",
    )
    parser.add_argument(
        "--model_predictions",
        default=False,
        action="store_true",
        help="Score predictions based on pre-train model",
    )
    parser.add_argument(
        "-o",
        "--output-folder",
        dest="output_folder",
        help="Specify the output folder for the summary files."
    )
    parser.add_argument(
        "--requant-mode",
        dest="requant_mode",
        help="Requantification mode (one of 'best', 'org' or 'fltr')."
    )
    parser.add_argument(
        "--model-pred-threshold",
        dest="model_pred_threshold",
        help="Fusion prediction model threshold."
    )
    parser.add_argument(
        "--fusion-tools",
        dest="fusion_tools",
        help="Fusion tools."
    )
    parser.set_defaults(func=summarize_data_command)


def summarize_data_command(args):
    summary = FusionSummary(
        input_fusions=args.input_fusions,
        input_fusion_context_seqs=args.input_fusion_context_seqs,
        input_requant_cpm=args.input_requant_cpm,
        input_requant_counts=args.input_requant_counts,
        input_reads_stats=args.input_reads_stats,
        input_model_pred=args.input_model_pred,
        output_folder=args.output_folder,
        requant_mode=args.requant_mode,
        model_pred_threshold=float(args.model_pred_threshold),
        fusion_tools=args.fusion_tools.split(","),
    )
    summary.run(args.model_predictions)



from configparser import NoOptionError, NoSectionError
from unittest import TestCase

import pkg_resources

import easy_fuse.tests
from easy_fuse.misc.config import EasyFuseConfiguration


class TestConfig(TestCase):
    def test_config_loading(self):
        config_file = pkg_resources.resource_filename(
            easy_fuse.tests.__name__, "resources/test.config.ini"
        )
        config = EasyFuseConfiguration(config_file=config_file)
        self.assertIsInstance(config, EasyFuseConfiguration)
        self.assertEqual(config.config_file, config_file)
        self.assertEqual(config.get("general", "pipeline_name"), "EasyFuse")
        self.assertEqual(config.get("general", "min_read_len_perc"), "0.75")
        self.assertEqual(config.get("general", "max_dist_proper_pair"), "200000")
        self.assertEqual(config.get("resources", "qc"), "6,10")
        self.assertEqual(config.get("commands", "fastqc"), "/path/to/fastqc_bin")
        self.assertEqual(
            config.get("references", "genome_fasta"),
            "/path/to/Homo_sapiens.GRCh38.dna.primary_assembly.fa",
        )
        self.assertEqual(config.get("indices", "star"), "/path/to/STAR_idx/")
        self.assertEqual(
            config.get("other_files", "infusion_cfg"),
            "/path/to/infusion_index/infusion.cfg",
        )

    def test_non_existing_value(self):
        config_file = pkg_resources.resource_filename(
            easy_fuse.tests.__name__, "resources/test.config.ini"
        )
        config = EasyFuseConfiguration(config_file=config_file)
        with self.assertRaises(NoOptionError):
            config.get("general", "idontexist")

    def test_non_existing_section(self):
        config_file = pkg_resources.resource_filename(
            easy_fuse.tests.__name__, "resources/test.config.ini"
        )
        config = EasyFuseConfiguration(config_file=config_file)
        with self.assertRaises(NoSectionError):
            config.get("idontexist", "pipeline_name")

    def test_non_existing_config_file(self):
        config_file = pkg_resources.resource_filename(
            easy_fuse.tests.__name__, "resources/idontexist.config.ini"
        )
        with self.assertRaises(AssertionError):
            EasyFuseConfiguration(config_file)



#!/usr/bin/env python3

"""
Simple method collection for:
1) Job submission to slurm workload manager (Creates an sbatch file and starts it)
2) Program execution outside the scheduler with "subprocess"
3) Slurm queue querying

@author: Tron (PASO), BNT (URLA)
@version: 20180118
"""

import os
import subprocess
import sys
from argparse import ArgumentParser

from logzero import logger


def get_jobs_by_name(name, system="slurm"):
    """This function returns a list of job ids for a given job name."""
    if system == "slurm":
        return get_jobs_by_name_slurm(name)
    elif system == "pbs":
        return get_jobs_by_name_pbs(name)
    else:
        return []


def get_jobs_by_name_pbs(name):
    jobs = []
    output = ""
    try:
        output = subprocess.check_output(["qstat", "-f"], universal_newlines=True)
    except:
        output = subprocess.Popen(
            ["qstat", "-f"], stdout=subprocess.PIPE
        ).communicate()[0]

    job_id = ""
    job_name = ""
    for x in output.rstrip().split("\n"):
        line = x.lstrip()
        if line.startswith("Job Id"):
            job_id = line.rsplit(" ", 1)[1]
        if line.startswith("Job_Name"):
            job_name = line.split(" = ")[1]

            if name == job_name:
                jobs.append(job_id)
            job_id = ""
            job_name = ""
    return jobs


def get_jobs_by_name_slurm(name):
    """Check if slurm job is already running (by name) and return its jobid if it does"""
    output = ""
    try:
        output = subprocess.check_output(
            ["squeue", "-o %i %j"], universal_newlines=True
        )
    except subprocess.CalledProcessError as call_error:
        logger.error(call_error.output)
        output = subprocess.Popen(
            ["squeue", "-o %i %j"], stdout=subprocess.PIPE
        ).communicate()[0]
    jobs = []
    for job_list in output.rstrip().split("\n")[1:]:
        elements = job_list.lstrip().split()

        if name in elements[1]:
            jobs.append(elements[0])
    return jobs


def submit(
    job_name,
    cmd,
    cores,
    mem_usage,
    output_results_folder,
    dependencies,
    partitions,
    userid,
    timelimit,
    mail,
    module_file,
    sched="slurm",
):
    if sched == "slurm":
        _submit_slurm(
            job_name,
            cmd,
            cores,
            mem_usage,
            output_results_folder,
            dependencies,
            partitions,
            userid,
            timelimit,
            mail,
            module_file,
        )
    else:
        _submit_nonqueue(job_name, cmd)


def _submit_nonqueue(job_name, cmd):
    #    if module_file:
    #        cmd = " && ".join(["source " + module_file, " ".join(cmd)]).split(" ")
    logger.info("Running {}".format(job_name))
    logger.info("CMD: {}".format(" ".join(cmd)))
    if ">" in cmd:
        index = cmd.index(">")
        output_file = cmd[index + 1]
        adj_cmd = cmd[:index]
        with open(output_file, "w") as fp:
            subprocess.run(adj_cmd, stdout=fp)
    else:
        p = subprocess.Popen(
            cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=False
        )
        (_, stderrdata) = p.communicate()
        r = p.returncode
        if r != 0:
            logger.error(
                'Error: Command "{}" returned non-zero exit status'.format(cmd)
            )
            logger.error(stderrdata)
            sys.exit(1)


def _submit_slurm(
    job_name,
    cmd,
    cores,
    mem_usage,
    output_results_folder,
    dependencies,
    partitions,
    userid,
    timelimit,
    mail,
    module_file,
):
    """This function submits a predefined job with specific SBATCH parameters to the Slurm workload manager system."""
    # add dependencies if necessary
    depend = "\n"
    if dependencies:
        depend = "#SBATCH --dependency=afterok:{}\n".format(":".join(dependencies))
    # add mail delivery
    mail_type = "\n"
    mail_user = "\n"
    if mail:
        mail_type = "#SBATCH --mail-type=END,FAIL,TIME_LIMIT_80\n"
        mail_user = "#SBATCH --mail-user={}\n".format(mail)
    # save stderr and stdout in different files
    error_file = os.path.join(output_results_folder, "error.log")
    output_file = os.path.join(output_results_folder, "output.log")

    # Generate sbatch script for your job in working dir
    slurm_script = os.path.join(output_results_folder, job_name + ".sbatch")
    with open(slurm_script, "w") as sbatch:
        sbatch.writelines(
            [
                "#!/bin/bash\n",
                "#SBATCH -J {}\n".format(job_name),
                "#SBATCH -p {}\n".format(partitions),
                "#SBATCH --account {}\n".format(userid),
                "#SBATCH --kill-on-invalid-dep=yes\n",
                "#SBATCH --cpus-per-task={}\n".format(cores),
                "#SBATCH --mem={}G\n".format(mem_usage),
                "#SBATCH --time={}\n".format(timelimit),
                depend,
                "#SBATCH -D {}\n".format(output_results_folder),
                "#SBATCH --error={}\n".format(error_file),
                "#SBATCH --output={}\n".format(output_file),
                mail_type,
                mail_user,
                "set -eo pipefail -o nounset\n",
                ". {}\n".format(module_file) if module_file else "\n",
                'srun echo "$(date) Slurm job started." > {0}.runtime && {1} && echo "$(date) Slurm job finished." >> {0}.runtime\n'.format(
                    job_name, cmd
                ),
            ]
        )
    # and run it
    try:
        output = subprocess.check_output(["sbatch", slurm_script])
        logger.info("{} ({})".format(output.rstrip(), job_name))
    except subprocess.CalledProcessError as call_error:
        logger.error(call_error.output)
        logger.error("Could not start slurm script {}".format(slurm_script))
        sys.exit(1)


def main():
    parser = ArgumentParser(description="Handle data streams")
    parser.add_argument(
        "-j", "--job-name", dest="job_name", help="Specify the job name.", required=True
    )
    parser.add_argument(
        "-s",
        "--script",
        dest="script",
        help="Specify the script to process.",
        required=True,
    )
    parser.add_argument(
        "-c",
        "--cores",
        dest="cores",
        type=int,
        help="Specify the number of cores to run the script with.",
        required=True,
    )
    parser.add_argument(
        "-m",
        "--memory",
        dest="memory",
        type=int,
        help="Specify the amount of memory to run your script with.",
        required=True,
    )
    parser.add_argument(
        "-o", "--output", dest="output", help="Select the output folder.", required=True
    )
    parser.add_argument(
        "-p",
        "--partitions",
        dest="partitions",
        help="Select the slurm partitions for this job.",
        default="allNodes",
    )
    parser.add_argument(
        "-u",
        "--userid",
        dest="userid",
        help="Select the slurm account to run the job on.",
        required=True,
    )
    parser.add_argument(
        "-t",
        "--timelimit",
        dest="timelimit",
        help="Select the timelimit for the job.",
        default="30-00:00:0",
    )
    args = parser.parse_args()

    job_name = args.job_name
    cmd = args.script
    cores = args.cores
    mem_usage = args.memory
    output_results_folder = args.output
    dependencies = ""
    submit(
        job_name,
        cmd,
        cores,
        mem_usage,
        output_results_folder,
        dependencies,
        args.partitions,
        args.userid,
        args.timelimit,
        "",
        "",
    )



#!/usr/bin/env python3

"""
Read input bam file generated by star
Star must have been run with "--outSAMmultNmax 1" and "--chimSegmentMin X"
Aligned pairs are filtered according to their likely of being chimeric with 4 subsequent filter:
    1) the pair is defined as chimeric by star
    2) at least one of the reads in the pair is unmapped
    3) the pair is not a proper pair (unexpected distance)
    4) 10 or more bases of at least one of the reads in the pair are soft-clipped
Multi-mappers are currently discarded as long as they are not belonging to filter "1"
All pairs that are not either "1-4", are defined as unlikely chimeric
A log file is generated after successful completion which contains information about:
    1) number of input vs filtered reads
    2) runtime/speed of the processing
    3) filter counts
    4) 2 qc tests for fast checking that counted pairs are as expected

@author: Tron (PASO), BNT (URLA)
@version: 20181126
"""
import os
import sys
import time

import logzero
import pysam  # pysam is not available for windows (where I run pylint) => pylint: disable=E0401
from logzero import logger

from easy_fuse.misc.count_input_reads import get_input_read_count_from_star


# pylint: disable=line-too-long
#         yes they are partially, but I do not consider this to be relevant here
class ReadSelection(object):
    """Select alignments belonging to putative fusions from an s/bam file"""

    def __init__(self, bam, output, context_file, input_reads_stats):
        """Parameter initialization"""
        self.bam_file = bam
        self.in_bam = pysam.AlignmentFile(bam, "rb")
        self.filtered_bam = pysam.AlignmentFile(output, "wb", template=self.in_bam)
        # a list of ints is used to count pairs assigned to different filters
        # 0: count_input_alignments
        # 1: count_input_pairs
        # 2: count_filtered_pairs
        # 3: count_multimapped
        # 4: count_star_chimeric_alignments
        # 5: count_qcd_alignments
        # 6: count_unmapped
        # 7: count_10bp_s_clip
        # 8: count_proper_pair
        # 9: count_not_filtered_but_in_fusion_gene
        self.counter = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
        self.last_time = 0
        logzero.logfile("{}.fusionReadFilterLog".format(output))

        # loads read counts
        assert os.path.exists(input_reads_stats) and os.path.isfile(input_reads_stats), \
            "Read stats file not found: {}".format(input_reads_stats)
        self.input_read_count = get_input_read_count_from_star(input_reads_stats)

        self.coord_dict = {}
        self.get_ranges(context_file)

    def get_ranges(self, context_file):
        """vlkfh"""
        with open(context_file, "r") as cfile:
            next(cfile)  # skip header
            for line in cfile:
                line_splitter = line.split(";")
                self.ranges_to_dict(line_splitter[-3])  # wt1 ranges
                self.ranges_to_dict(line_splitter[-2])  # wt2 ranges
        self.merge_ranges()

    def ranges_to_dict(self, context_range):
        """stub"""
        chrom, start, stop = context_range.rstrip().split(":")
        if not chrom in self.coord_dict:
            self.coord_dict[chrom] = []
        self.coord_dict[chrom].append([int(start), int(stop)])

    def merge_ranges(self):
        """stub"""
        for chrom in self.coord_dict:
            chr_coords = self.coord_dict[chrom]
            chr_coords.sort(key=lambda start: start[0])
            merged_coords = [chr_coords[0]]
            for current in chr_coords:
                previous = merged_coords[-1]
                if current[0] <= previous[1]:
                    previous[1] = max(previous[1], current[1])
                else:
                    merged_coords.append(current)
            self.coord_dict[chrom] = merged_coords

    def check_overlap(self, read1, read2, read1_flag, read2_flag, member_count):
        """Classify a read pairs into different groups"""
        self.counter[1] += 1
        # count star chimeric
        if member_count > 2:
            self.counter[4] += 1
            return True
        # count multimapping reads
        if read1.get_tag("NH") > 1 or read2.get_tag("NH") > 1:
            self.counter[3] += 1
            return False
        # count pairs containing unmapped mates
        if read1_flag & 0x4 or read2_flag & 0x4:
            self.counter[6] += 1
            return True
        # check if the pair is a "proper pair" (both mapped, adequat distance)
        if not read1_flag & 0x2 or not read2_flag & 0x2:
            self.counter[8] += 1
            return True
        # check if 10 (default) or more soft-clippings (S) are in the alignment
        if (
            read1.query_alignment_length + 10 <= read1.query_length
            or read2.query_alignment_length + 10 <= read2.query_length
        ):
            self.counter[7] += 1
            return True

        # now check whether one of the reads is overlapping
        if self.get_region_overlap(read1) or self.get_region_overlap(read2):
            self.counter[9] += 1
            return True
        return False

    def get_region_overlap(self, read):
        """bla"""
        try:
            for (region_start, region_stop) in self.coord_dict[read.reference_name]:
                if (
                    region_start <= read.reference_start <= region_stop
                    or region_start <= read.reference_end <= region_stop
                ):
                    return True
        except KeyError:
            # A key error indicates that the read maps to a chromosome which contains no region of interest.
            # This is obviously equivalent to "not an overlap"
            return False
        return False

    def run(self):
        """Walk linewise through a s/bam file and send proper read pairs to classification"""
        logger.info("Starting fusion read filtering")
        read1 = read2 = None
        self.last_time = time.time()
        last_query = ""
        count_current_query_member = 0
        # read filtering works as follows:
        # urla: I wrote the processing in a way, that implementing the analyses of multi-mapping reads should be straight forward
        #       For this, however, one would need to check in the group of multi-mapping alignments whether there is no possible "normal mapping"
        #       I could so far think of two possible solutions: (1) running "classify_pair" on all possible combinations of the multimapping pairs,
        #       until a "normal mapping" was found. If this is never the case, the pair is written to the filtered out. This, however, will have
        #       a strong effect on runtimes! (2) with the chimeric multi-mapping setting in the most recent star versions. This should actually
        #       work directly as star chimeric classification outranks multi-mapping disposal. It has, however, not been throughly evaluated by the star community
        for read in self.in_bam.fetch(
            until_eof=True
        ):  # iterate through alignments as they appear in the file (this is mandatory becaue (a) we cannot create an index, (b) want to include unmapped reads and (c) have many references in the header during 2nd filtering)
            self.counter[0] += 1
            read_flag = read.flag

            if last_query != read.query_name and self.counter[0] > 1:
                if self.check_overlap(
                    read1, read2, read1_flag, read2_flag, count_current_query_member
                ):
                    self.filtered_bam.write(read1)
                    self.filtered_bam.write(read2)
                    self.counter[2] += 1
                read1 = read2 = None
                read1_flag = read2_flag = None
                count_current_query_member = 0

            count_current_query_member += 1
            # ignore all alignments which are either supplemental, vendor qc'd, secondary or duplicates
            # urla: duplicate flagging/removal prior to running this script would most probably lead to several errors
            #       Nevertheless, it is a very fast check and I would also not recommend that someone does deduplication on rna-seq data!
            if read_flag > 255:
                self.counter[5] += 1
            else:
                # urla: the following commented version should be fine, but the other one is still kept in order to be more error-aware
                #                if read.is_read1:
                #                    read1 = read
                #                else:
                #                    read2 = read
                if not read1 and read.is_read1:
                    read1 = read
                    read1_flag = read_flag
                elif not read2 and read.is_read2:
                    read2 = read
                    read2_flag = read_flag
                else:
                    logger.error(
                        "Neither r1 nor r2??? Read: {0}; R1: {1}; R2: {2}; bamLine: {3}".format(
                            read, read1, read2, self.counter[0]
                        )
                    )
                    sys.exit(1)
            # urla: uncomment the following, if you'd like to have stats updates during the run and not only at the end
            if self.counter[0] % 1000000 == 0:
                self.print_stats()
            last_query = read.query_name

        # once EOF is reached, the very last pair has to be classified additionally
        if self.check_overlap(
            read1, read2, read1_flag, read2_flag, count_current_query_member
        ):
            self.filtered_bam.write(read1)
            self.filtered_bam.write(read2)
            self.counter[2] += 1

        # close reading/writing stream
        self.in_bam.close()
        self.filtered_bam.close()
        self.print_stats()
        logger.info("Finished fusion read filtering")

    def print_stats(self):
        """print collected statistics to the log file"""
        this_time = time.time()
        time_taken = this_time - self.last_time
        time_taken_1m = float(time_taken * 1000000) / float(self.counter[0])
        self.last_time = this_time
        logger.info(
            "Processed {0} alignments, {1} of {2} pairs remained after filtering ({3:.2%}) ({4:.2f}s / 1M alignments; {5:.2f}s in total)".format(
                self.counter[0],
                self.counter[2],
                self.counter[1],
                float(self.counter[2]) / float(self.counter[1]),
                time_taken_1m,
                time_taken,
            )
        )

        qc1 = False
        if self.input_read_count == self.counter[1]:
            qc1 = True
        qc2 = False
        if (self.counter[4] == self.counter[5]) and (
            (self.counter[0] - self.counter[5]) * 0.5 == self.counter[1]
        ):
            qc2 = True

        # 0: count_input_alignments
        # 1: count_input_pairs
        # 2: count_filtered_pairs
        # 3: count_multimapped
        # 4: count_star_chimeric_alignments
        # 5: count_qcd_alignments
        # 6: count_unmapped
        # 7: count_10bp_s_clip
        # 8: count_proper_pair
        logger.info(
            "Star_chimeric (chim alignment from star):\t{} pairs (filtered)".format(
                self.counter[4]
            )
        )
        logger.info(
            "QC'd (additional Star_chimeric alignment):\t{} alignments (included in above)".format(
                self.counter[5]
            )
        )
        logger.info(
            "Multimapped (1 < x <= 100 equal mappings):\t{} pairs (discarded)".format(
                self.counter[3]
            )
        )
        logger.info(
            "Unmapped (no mapping or >100 multi map):\t{} pairs (filtered)".format(
                self.counter[6]
            )
        )
        logger.info(
            "No proper pair (unexpected read distance):\t{} pairs (filtered)".format(
                self.counter[8]
            )
        )
        logger.info(
            "10bp_s_clip (>9bp soft-clipped in cigar):\t{} pairs (filtered)".format(
                self.counter[7]
            )
        )
        logger.info(
            'Unlikely chimeric ("normal" mappings): \t{} pairs (discarded)'.format(
                self.counter[1]
                - self.counter[4]
                - self.counter[3]
                - self.counter[6]
                - self.counter[8]
                - self.counter[7]
            )
        )
        logger.info(
            "Not filtered, but mapped to fusion gene:\t{} pairs (filtered)".format(
                self.counter[9]
            )
        )
        logger.info("Filter QC1 (fq reads = bam alignments):\t{}".format(qc1))
        logger.info("Filter QC2 (QC'd alignments are chimeric):\t{}".format(qc2))



def add_read_selection_args(parser):
    """Add read selection arguments to parser"""
    parser.add_argument("-i", "--input", dest="input", help="Specify input BAM file. Must be indexed and sorted by coordinate.")
    parser.add_argument(
        "-i2", "--input2", dest="input2", help="Specify context_seq file file"
    )
    parser.add_argument(
        "--input-reads-stats",
        dest="input_reads_stats",
        required=True,
        help="Path to input file with reads stats",
    )
    parser.add_argument("-o", "--output", dest="output", help="Specify output prefix")
    parser.set_defaults(func=read_selection_command)


def read_selection_command(args):
    """Run read selection"""
    stats = ReadSelection(args.input, args.output, args.input2, args.input_reads_stats)
    stats.run()
